import tomlkit
import yaml
from typing import Any, Callable, Dict, MutableMapping, Sequence, Optional, Tuple, Type

from dlt.common.configuration.utils import auto_cast, auto_config_fragment
from dlt.common.utils import update_dict_nested

from .provider import ConfigProvider, get_key_name


class BaseDocProvider(ConfigProvider):
    _config_doc: Dict[str, Any]
    """Holds a dict with config values"""

    def __init__(self, config_doc: Dict[str, Any]) -> None:
        self._config_doc = config_doc

    @staticmethod
    def get_key_name(key: str, *sections: str) -> str:
        return get_key_name(key, ".", *sections)

    def get_value(
        self, key: str, hint: Type[Any], pipeline_name: str, *sections: str
    ) -> Tuple[Optional[Any], str]:
        full_path = sections + (key,)
        if pipeline_name:
            full_path = (pipeline_name,) + full_path
        full_key = self.get_key_name(key, pipeline_name, *sections)
        node = self._config_doc
        try:
            for k in full_path:
                if not isinstance(node, dict):
                    raise KeyError(k)
                node = node[k]
            return node, full_key
        except KeyError:
            return None, full_key

    def set_value(self, key: str, value: Any, pipeline_name: Optional[str], *sections: str) -> None:
        """Sets `value` under `key` in `sections` and optionally for `pipeline_name`

        If key already has value of type dict and value to set is also of type dict, the new value
        is merged with old value.
        """
        self._set_value(self._config_doc, key, value, pipeline_name, *sections)

    def set_fragment(
        self, key: Optional[str], value_or_fragment: str, pipeline_name: str, *sections: str
    ) -> None:
        """Tries to interpret `value_or_fragment` as a fragment of toml, yaml or json string and replace/merge into config doc.

        If `key` is not provided, fragment is considered a full document and will replace internal config doc. Otherwise
        fragment is merged with config doc from the root element and not from the element under `key`!

        For simple values it falls back to `set_value` method.
        """
        self._config_doc = self._set_fragment(
            self._config_doc, key, value_or_fragment, pipeline_name, *sections
        )

    def to_toml(self) -> str:
        return tomlkit.dumps(self._config_doc)

    def to_yaml(self) -> str:
        return yaml.dump(
            self._config_doc, allow_unicode=True, default_flow_style=False, sort_keys=False
        )

    @property
    def supports_sections(self) -> bool:
        return True

    @property
    def is_empty(self) -> bool:
        return len(self._config_doc) == 0

    @staticmethod
    def _set_value(
        master: MutableMapping[str, Any],
        key: str,
        value: Any,
        pipeline_name: Optional[str],
        *sections: str,
    ) -> None:
        if pipeline_name:
            sections = (pipeline_name,) + sections
        if key is None:
            raise ValueError("dlt_secrets_toml must contain toml document")

        # descend from root, create tables if necessary
        for k in sections:
            if not isinstance(master, dict):
                raise KeyError(k)
            if k not in master:
                master[k] = {}
            master = master[k]
        if isinstance(value, dict):
            # remove none values, TODO: we need recursive None removal
            value = {k: v for k, v in value.items() if v is not None}
            # if target is also dict then merge recursively
            if isinstance(master.get(key), dict):
                update_dict_nested(master[key], value)
                return
        master[key] = value

    @staticmethod
    def _set_fragment(
        master: MutableMapping[str, Any],
        key: Optional[str],
        value_or_fragment: str,
        pipeline_name: str,
        *sections: str,
    ) -> Any:
        """Tries to interpret `value_or_fragment` as a fragment of toml, yaml or json string and replace/merge into config doc.

        If `key` is not provided, fragment is considered a full document and will replace internal config doc. Otherwise
        fragment is merged with config doc from the root element and not from the element under `key`!

        For simple values it falls back to `set_value` method.
        """
        fragment = auto_config_fragment(value_or_fragment)
        if fragment is not None:
            # always update the top document
            if key is None:
                master = fragment
            else:
                # TODO: verify that value contains only the elements under key
                update_dict_nested(master, fragment)
        else:
            # set value using auto_cast
            BaseDocProvider._set_value(
                master, key, auto_cast(value_or_fragment), pipeline_name, *sections
            )
        return master


class CustomLoaderDocProvider(BaseDocProvider):
    def __init__(
        self,
        name: str,
        loader: Callable[[], Dict[str, Any]],
        supports_secrets: bool = True,
        locations: Sequence[str] = None,
    ) -> None:
        """Provider that calls `loader` function to get a Python dict with config/secret values to be queried.
        The `loader` function typically loads a string (ie. from file), parses it (ie. as toml or yaml), does additional
        processing and returns a Python dict to be queried.

        Instance of CustomLoaderDocProvider must be registered for the returned dict to be used to resolve config values.
        >>> import dlt
        >>> dlt.config.register_provider(provider)

        Args:
            name(str): name of the provider that will be visible ie. in exceptions
            loader(Callable[[], Dict[str, Any]]): user-supplied function that will load the document with config/secret values
            supports_secrets(bool): allows to store secret values in this provider
            locations(Sequence[str], optional): a human readable list of locations from which configuration is read, used to
                display meaningful error messages when configurations was not resolved.

        """
        self._name = name
        self._supports_secrets = supports_secrets
        self._locations = locations
        super().__init__(loader())

    @property
    def name(self) -> str:
        return self._name

    @property
    def supports_secrets(self) -> bool:
        return self._supports_secrets

    @property
    def is_writable(self) -> bool:
        return True

    @property
    def locations(self) -> Sequence[str]:
        return self._locations



"""Module with built in sources and source building blocks"""
from dlt.common.typing import TDataItem, TDataItems
from dlt.extract import DltSource, DltResource, Incremental as incremental
from dlt.extract.reference import AnySourceFactory, SourceReference, UnknownSourceReference
from . import credentials, config


__all__ = [
    "DltSource",
    "DltResource",
    "SourceReference",
    "UnknownSourceReference",
    "AnySourceFactory",
    "TDataItem",
    "TDataItems",
    "incremental",
    "credentials",
    "config",
]



"""Transactional file system operations.

The lock implementation allows for multiple readers and a single writer.
It can be used to operate on a single file atomically both locally and on
cloud storage. The lock can be used to operate on entire directories by
creating a lock file that resolves to an agreed upon path across processes.
"""
import random
import string
import time
import typing as t
from pathlib import Path
import posixpath
from contextlib import contextmanager
from threading import Timer
import fsspec

from dlt.common.pendulum import pendulum, timedelta
from dlt.common.storages.fsspec_filesystem import MTIME_DISPATCH


def lock_id(k: int = 4) -> str:
    """Generate a time based random id.

    Args:
        k: The length of the suffix after the timestamp.

    Returns:
        A time sortable uuid.
    """
    suffix = "".join(random.choices(string.ascii_lowercase, k=k))
    return f"{time.time_ns()}{suffix}"


class Heartbeat(Timer):
    """A thread designed to periodically execute a fn."""

    daemon = True

    def run(self) -> None:
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)
        self.finished.set()


class TransactionalFile:
    """A transaction handler which wraps a file path."""

    POLLING_INTERVAL = 0.5
    LOCK_TTL_SECONDS = 30.0

    def __init__(self, path: str, fs: fsspec.AbstractFileSystem) -> None:
        """Creates a new FileTransactionHandler.

        Args:
            path: The path to lock.
            fs: The fsspec file system.
        """
        proto = fs.protocol[0] if isinstance(fs.protocol, (list, tuple)) else fs.protocol
        self.extract_mtime = MTIME_DISPATCH.get(proto, MTIME_DISPATCH["file"])

        parsed_path = Path(path)
        if not parsed_path.is_absolute():
            raise ValueError(
                f"`TransactionalFile` requires an absolute path. Received path: `{path}`"
            )
        self.path = path
        if proto == "file":
            # standardize path separator to POSIX. fsspec always uses POSIX. Windows may use either.
            self.path = parsed_path.as_posix()

        self.lock_prefix = f"{self.path}.lock"

        self._fs = fs
        self._original_contents: t.Optional[bytes] = None
        self._is_locked = False
        self._heartbeat: t.Optional[Heartbeat] = None

    def _start_heartbeat(self) -> Heartbeat:
        """Create a thread that will periodically update the mtime."""
        self._stop_heartbeat()
        self._heartbeat = Heartbeat(
            TransactionalFile.LOCK_TTL_SECONDS / 2,
            self._fs.touch,
            args=(self.lock_path,),
        )
        self._heartbeat.start()
        return self._heartbeat

    def _stop_heartbeat(self) -> None:
        """Stop the heartbeat thread if it exists."""
        if self._heartbeat is not None:
            self._heartbeat.cancel()
            self._heartbeat = None

    def _sync_locks(self) -> t.List[str]:
        """Gets a list of lock names after removing stale locks. The list is time-sortable with earliest created lock coming first."""
        output = []
        now = pendulum.now()

        for lock in self._fs.ls(posixpath.dirname(self.lock_path), refresh=True, detail=True):
            name = lock["name"]
            if not name.startswith(self.lock_prefix):
                continue
            # Purge stale locks
            mtime = self.extract_mtime(lock)
            if now - mtime > timedelta(seconds=TransactionalFile.LOCK_TTL_SECONDS):
                try:  # Janitors can race, so we ignore errors
                    self._fs.rm(name)
                except OSError:
                    pass
                continue
            # The name is timestamp + random suffix and is time sortable
            output.append(name)
        if not output:
            raise RuntimeError(
                f"Lock syncing failed. No lock file found for path `{self.path}` and lock"
                f" `{self.lock_path}`"
            )
        return output

    def read(self) -> t.Optional[bytes]:
        """Reads data from the file if it exists."""
        if self._fs.isfile(self.path):
            return t.cast(bytes, self._fs.cat(self.path))
        return None

    def write(self, content: bytes) -> None:
        """Writes data within the transaction."""
        if not self._is_locked:
            raise RuntimeError("Cannot write to a file without a lock.")
        if self._fs.isdir(self.path):
            raise RuntimeError("Cannot write to a directory.")
        self._fs.pipe(self.path, content)

    def rollback(self) -> None:
        """Rolls back the transaction."""
        if not self._is_locked:
            raise RuntimeError("Cannot rollback without a lock.")
        if self._original_contents is not None:
            self.write(self._original_contents)
        elif self._fs.isfile(self.path):
            self._fs.rm(self.path)

    def acquire_lock(
        self, blocking: bool = True, timeout: float = -1, jitter_mean: float = 0
    ) -> bool:
        """Acquires a lock on a path. Mimics the stdlib's `threading.Lock` interface.

        Acquire a lock, blocking or non-blocking.

        When invoked with the blocking argument set to True (the default), block until
        the lock is unlocked, then set it to locked and return True.

        When invoked with the blocking argument set to False, do not block. If a call
        with blocking set to True would block, return False immediately; otherwise, set
        the lock to locked and return True.

        When invoked with the floating-point timeout argument set to a positive value,
        block for at most the number of seconds specified by timeout and as long as the
        lock cannot be acquired. A timeout argument of -1 specifies an unbounded wait.
        If blocking is False, timeout is ignored. The stdlib would raise a ValueError.

        If you expect a large concurrency on the lock, you can set the jitter_mean to a
        value > 0. This will introduce a short random gap before locking procedure
        starts.

        The return value is True if the lock is acquired successfully, False if
        not (for example if the timeout expired).
        """
        if self._is_locked:
            return True

        if jitter_mean > 0:
            time.sleep(random.random() * jitter_mean)  # Add jitter to avoid thundering herd
        self.lock_path = f"{self.lock_prefix}.{lock_id()}"
        self._fs.touch(self.lock_path)
        locks = self._sync_locks()
        active_lock = min(locks)
        start = time.monotonic()

        while active_lock != self.lock_path:
            if not blocking or (timeout > 0 and time.monotonic() - start > timeout):
                self._fs.rm(self.lock_path)
                return False

            time.sleep(random.random() + TransactionalFile.POLLING_INTERVAL)
            locks = self._sync_locks()
            if self.lock_path not in locks:
                self._fs.touch(self.lock_path)
                locks = self._sync_locks()

            active_lock = min(locks)

        self._original_contents = self.read()
        self._is_locked = True
        self._start_heartbeat()
        return True

    def release_lock(self) -> None:
        """Releases a lock on a key.

        This is idempotent and safe to call multiple times.
        """
        if self._is_locked:
            self._stop_heartbeat()
            self._fs.rm(self.lock_path)
            self._is_locked = False
            self._original_contents = None

    @contextmanager
    def lock(self, timeout: t.Optional[float] = None) -> t.Iterator[None]:
        """A context manager that acquires and releases a lock on a path.

        This is a convenience method for acquiring a lock and reading the contents of
        the file. It will release the lock when the context manager exits. This is
        useful for reading a file and then writing it back out as a transaction. If
        the lock cannot be acquired, this will raise a RuntimeError. If the lock is
        acquired, the contents of the file will be returned. If the file does not
        exist, None will be returned. If an exception is raised within the context
        manager, the transaction will be rolled back.

        Args:
            timeout: The timeout for acquiring the lock. If None, this will use the
                default timeout. A timeout of -1 will wait indefinitely.

        Raises:
            RuntimeError: If the lock cannot be acquired.
        """
        if timeout is None:
            timeout = TransactionalFile.LOCK_TTL_SECONDS + 1
        if not self.acquire_lock(timeout=timeout):
            raise RuntimeError("Could not acquire lock.")
        try:
            yield
        except Exception:
            self.rollback()
            raise
        finally:
            self.release_lock()

    def __del__(self) -> None:
        """Stop the heartbeat thread on gc. Locks should be released explicitly."""
        self._stop_heartbeat()



import base64
from abc import abstractmethod, ABC
from functools import lru_cache
import math
import hashlib
import re
from typing import Sequence, ClassVar


RE_LEADING_UNDERSCORES = re.compile(r"^_+")


class NamingConvention(ABC):
    """Initializes naming convention to generate identifier with `max_length` if specified. Base naming convention
    is case sensitive by default
    """

    _TR_TABLE: ClassVar[bytes] = bytes.maketrans(b"/+", b"ab")
    _DEFAULT_COLLISION_PROB: ClassVar[float] = 0.001
    PATH_SEPARATOR: ClassVar[str] = "__"
    """Subsequent nested fields will be separated with the string below, applies both to field and table names"""

    def __init__(self, max_length: int = None) -> None:
        self.max_length = max_length

    @property
    @abstractmethod
    def is_case_sensitive(self) -> bool:
        """Tells if given naming convention is producing case insensitive or case sensitive identifiers."""
        pass

    @abstractmethod
    def normalize_identifier(self, identifier: str) -> str:
        """Normalizes and shortens the identifier according to naming convention in this function code"""
        if identifier is None:
            raise ValueError("`name` is None")
        identifier = identifier.strip()
        if not identifier:
            raise ValueError(identifier)
        return identifier

    def normalize_table_identifier(self, identifier: str) -> str:
        """Normalizes and shortens identifier that will function as a dataset, table or schema name, defaults to `normalize_identifier`"""
        return self.normalize_identifier(identifier)

    def make_path(self, *identifiers: str) -> str:
        """Builds path out of identifiers. Identifiers are neither normalized nor shortened"""
        return self.PATH_SEPARATOR.join(filter(lambda x: x.strip(), identifiers))

    def break_path(self, path: str) -> Sequence[str]:
        """Breaks path into sequence of identifiers"""
        # TODO: this is no longer needed if we modify all naming convention to do not contract
        #   underscores then also normalize_path will not be needed
        return [ident for ident in path.split(self.PATH_SEPARATOR) if ident.strip()]

    def normalize_path(self, path: str) -> str:
        """Breaks path into identifiers, normalizes components, reconstitutes and shortens the path"""
        normalized_idents = [self.normalize_identifier(ident) for ident in self.break_path(path)]
        # shorten the whole path
        return self.shorten_identifier(self.make_path(*normalized_idents), path, self.max_length)

    def normalize_tables_path(self, path: str) -> str:
        """Breaks path of table identifiers, normalizes components, reconstitutes and shortens the path"""
        normalized_idents = [
            self.normalize_table_identifier(ident) for ident in self.break_path(path)
        ]
        # shorten the whole path
        return self.shorten_identifier(self.make_path(*normalized_idents), path, self.max_length)

    def shorten_fragments(self, *normalized_idents: str) -> str:
        """Reconstitutes and shortens the path of normalized identifiers"""
        if not normalized_idents:
            return None
        path_str = self.make_path(*normalized_idents)
        return self.shorten_identifier(path_str, path_str, self.max_length)

    @classmethod
    def name(cls) -> str:
        """Naming convention name is the name of the module in which NamingConvention is defined"""
        if cls.__module__.startswith("dlt.common.normalizers.naming."):
            # return last component
            return cls.__module__.split(".")[-1]
        return cls.__module__

    def __str__(self) -> str:
        name = self.name()
        name += "_cs" if self.is_case_sensitive else "_ci"
        if self.max_length:
            name += f"_{self.max_length}"
        return name

    @staticmethod
    @lru_cache(maxsize=None)
    def shorten_identifier(
        normalized_ident: str,
        identifier: str,
        max_length: int,
        collision_prob: float = _DEFAULT_COLLISION_PROB,
    ) -> str:
        """Shortens the `name` to `max_length` and adds a tag to it to make it unique. Tag may be placed in the middle or at the end"""
        if max_length and len(normalized_ident) > max_length:
            # use original identifier to compute tag
            tag = NamingConvention._compute_tag(identifier, collision_prob)
            normalized_ident = NamingConvention._trim_and_tag(normalized_ident, tag, max_length)

        return normalized_ident

    @staticmethod
    def _compute_tag(identifier: str, collision_prob: float) -> str:
        # assume that shake_128 has perfect collision resistance 2^N/2 then collision prob is 1/resistance: prob = 1/2^N/2, solving for prob
        # take into account that we are case insensitive in base64 so we need ~1.5x more bits (2+1)
        tl_bytes = int(((2 + 1) * math.log2(1 / (collision_prob)) // 8) + 1)
        tag = (
            base64.b64encode(hashlib.shake_128(identifier.encode("utf-8")).digest(tl_bytes))
            .rstrip(b"=")
            .translate(NamingConvention._TR_TABLE)
            .lower()
            .decode("ascii")
        )
        return tag

    @staticmethod
    def _trim_and_tag(identifier: str, tag: str, max_length: int) -> str:
        assert len(tag) <= max_length
        remaining_length = max_length - len(tag)
        remaining_overflow = remaining_length % 2
        identifier = (
            identifier[: remaining_length // 2 + remaining_overflow]
            + tag
            + identifier[len(identifier) - remaining_length // 2 :]
        )
        assert len(identifier) == max_length
        return identifier

    @staticmethod
    def _remove_leading_underscores(identifier: str) -> str:
        return RE_LEADING_UNDERSCORES.sub("", identifier)



from dlt.common.pendulum import pendulum

RANDOM_DATE = pendulum.datetime(1768, 7, 21, 2, 56, 7, 3, tz="UTC")
TYPE_EXAMPLES = {
    "text": "<configure me>",
    "timestamp": RANDOM_DATE.to_iso8601_string(),
    "date": RANDOM_DATE.to_date_string(),
}



from packaging.version import Version
from typing import Any, Optional, Sequence, Type, Union, Dict, TYPE_CHECKING

from dlt.common import logger
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.data_writers.escape import escape_postgres_identifier, escape_duckdb_literal
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import TColumnSchema, TColumnType, TLoaderMergeStrategy, TTableSchema
from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.duckdb.configuration import DuckDbCredentials, DuckDbClientConfiguration

if TYPE_CHECKING:
    from duckdb import DuckDBPyConnection
    from dlt.destinations.impl.duckdb.duck import DuckDbClient
else:
    DuckDBPyConnection = Any


class DuckDbTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "JSON",
        "text": "VARCHAR",
        "double": "DOUBLE",
        "bool": "BOOLEAN",
        "date": "DATE",
        # Duck does not allow specifying precision on timestamp with tz
        "timestamp": "TIMESTAMP WITH TIME ZONE",
        "bigint": "BIGINT",
        "binary": "BLOB",
        "time": "TIME",
    }

    sct_to_dbt = {
        # VARCHAR(n) is alias for VARCHAR in duckdb
        # "text": "VARCHAR(%i)",
        "decimal": "DECIMAL(%i,%i)",
        "wei": "DECIMAL(%i,%i)",
    }

    dbt_to_sct = {
        "VARCHAR": "text",
        "JSON": "json",
        "DOUBLE": "double",
        "BOOLEAN": "bool",
        "DATE": "date",
        "TIMESTAMP WITH TIME ZONE": "timestamp",
        "BLOB": "binary",
        "DECIMAL": "decimal",
        "TIME": "time",
        # Int types
        "TINYINT": "bigint",
        "SMALLINT": "bigint",
        "INTEGER": "bigint",
        "BIGINT": "bigint",
        "HUGEINT": "bigint",
        "TIMESTAMP": "timestamp",
        "TIMESTAMP_S": "timestamp",
        "TIMESTAMP_MS": "timestamp",
        "TIMESTAMP_NS": "timestamp",
    }

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        precision = column.get("precision")
        if precision is None:
            return "BIGINT"
        # Precision is number of bits
        if precision <= 8:
            return "TINYINT"
        elif precision <= 16:
            return "SMALLINT"
        elif precision <= 32:
            return "INTEGER"
        elif precision <= 64:
            return "BIGINT"
        elif precision <= 128:
            return "HUGEINT"
        raise TerminalValueError(
            f"bigint with `{precision=:}` can't be mapped to DuckDB integer type"
        )

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        column_name = column["name"]
        table_name = table["name"]
        timezone = column.get("timezone", True)
        precision = column.get("precision", self.capabilities.timestamp_precision)

        if timezone and precision != self.capabilities.timestamp_precision:
            logger.warn(
                f"DuckDB does not support both timezone and precision for column '{column_name}' in"
                f" table '{table_name}'. Will default to timezone. Please set timezone to False to"
                " use precision types."
            )

        if timezone:
            # default timestamp mapping for timezone
            return None

        if precision == 0:
            return "TIMESTAMP_S"
        elif precision <= 3:
            return "TIMESTAMP_MS"
        elif precision <= 6:
            return "TIMESTAMP"
        elif precision <= 9:
            return "TIMESTAMP_NS"

        raise TerminalValueError(
            f"DuckDB doesn't support `{precision=:}` for datetime column `{column_name}` in table"
            f" `{table_name}`"
        )

    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        # duckdb provides the types with scale and precision
        db_type = db_type.split("(")[0].upper()
        if db_type == "DECIMAL":
            if precision == 38 and scale == 0:
                return dict(data_type="wei", precision=precision, scale=scale)
        return super().from_destination_type(db_type, precision, scale)


def duckdb_merge_strategies_selector(
    supported_merge_strategies: Sequence[TLoaderMergeStrategy],
    /,
    *,
    table_schema: TTableSchema,
) -> Sequence[TLoaderMergeStrategy]:
    try:
        import duckdb as _duckdb

        if Version(_duckdb.__version__) < Version("1.4.0"):
            legacy_strategies = list(supported_merge_strategies)
            legacy_strategies.remove("upsert")
            supported_merge_strategies = legacy_strategies
    except ImportError:
        # return default if duckdb not installed
        pass

    return supported_merge_strategies


def _set_duckdb_raw_capabilities(
    caps: DestinationCapabilitiesContext,
) -> DestinationCapabilitiesContext:
    caps.preferred_loader_file_format = "insert_values"
    caps.supported_loader_file_formats = ["insert_values", "parquet", "jsonl", "model"]
    caps.preferred_staging_file_format = None  # TODO remove because is default value
    caps.supported_staging_file_formats = []  # TODO is this different from default `None`?
    caps.type_mapper = DuckDbTypeMapper
    caps.escape_identifier = escape_postgres_identifier
    # all identifiers are case insensitive but are stored as is
    caps.escape_literal = escape_duckdb_literal
    caps.has_case_sensitive_identifiers = False
    caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
    caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
    caps.max_identifier_length = 65536
    caps.max_column_identifier_length = 65536
    caps.max_query_length = 32 * 1024 * 1024
    caps.is_max_query_length_in_bytes = True
    caps.max_text_data_type_length = 1024 * 1024 * 1024
    caps.is_max_text_data_type_length_in_bytes = True
    caps.supports_ddl_transactions = True
    caps.alter_add_multi_column = False
    caps.supports_truncate_command = False
    caps.supported_merge_strategies = ["delete-insert", "upsert", "scd2"]
    caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]
    caps.merge_strategies_selector = duckdb_merge_strategies_selector
    caps.sqlglot_dialect = "duckdb"
    caps.timestamp_precision = 6
    caps.max_timestamp_precision = 9  # nanosecond precision supported

    return caps


class duckdb(Destination[DuckDbClientConfiguration, "DuckDbClient"]):
    spec = DuckDbClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps = _set_duckdb_raw_capabilities(caps)
        return caps

    @property
    def client_class(self) -> Type["DuckDbClient"]:
        from dlt.destinations.impl.duckdb.duck import DuckDbClient

        return DuckDbClient

    def __init__(
        self,
        credentials: Union[DuckDbCredentials, Dict[str, Any], str, DuckDBPyConnection] = None,
        create_indexes: bool = False,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the DuckDB destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[DuckDbCredentials, Dict[str, Any], str, DuckDBPyConnection], optional): Credentials to connect to the duckdb database. Can be an instance of `DuckDbCredentials` or
                a path to a database file. Use :pipeline: to create a duckdb in the working folder of the pipeline.
                Instance of `DuckDbCredentials` allows to pass extensions, configs and pragmas to be set up for connection.
            create_indexes (bool, optional): Should unique indexes be created, defaults to False
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """

        super().__init__(
            credentials=credentials,
            create_indexes=create_indexes,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


duckdb.register()



from typing import Optional, Tuple, TypeVar, Generic, Type, Union, Any, List

from dlt.common import logger
from dlt.common.exceptions import MissingDependencyException
from dlt.common.schema.schema import Schema
from dlt.common.typing import TDataItems
from dlt.common.schema.exceptions import DataValidationError
from dlt.common.schema.typing import (
    TAnySchemaColumns,
    TSchemaContract,
    TSchemaEvolutionMode,
    TTableSchema,
)

try:
    from pydantic import BaseModel as PydanticBaseModel
    from dlt.common.libs.pydantic import (
        ValidationError,
        validate_and_filter_item,
        validate_and_filter_items,
    )
except (ModuleNotFoundError, MissingDependencyException):
    PydanticBaseModel = Any  # type: ignore[misc, assignment]

from dlt.extract.utils import get_data_item_format
from dlt.extract.items import TTableHintTemplate
from dlt.extract.items_transform import BaseItemTransform, ValidateItem


_TPydanticModel = TypeVar("_TPydanticModel", bound=PydanticBaseModel)


class PydanticValidator(ValidateItem, Generic[_TPydanticModel]):
    model: Type[_TPydanticModel]

    def __init__(
        self,
        model: Type[_TPydanticModel],
        column_mode: TSchemaEvolutionMode,
        data_mode: TSchemaEvolutionMode,
    ) -> None:
        from dlt.common.libs.pydantic import (
            _build_discriminator_map,
            apply_schema_contract_to_model,
            create_list_model,
        )

        BaseItemTransform.__init__(self)
        self.column_mode: TSchemaEvolutionMode = column_mode
        self.data_mode: TSchemaEvolutionMode = data_mode
        self.original_model = model
        self._discriminator_map = _build_discriminator_map(model)
        self.model = apply_schema_contract_to_model(model, column_mode, data_mode)
        self.list_model = create_list_model(self.model, column_mode, data_mode)

    def __call__(self, item: TDataItems, meta: Any = None) -> TDataItems:
        """Validate a data item against the pydantic model"""
        if item is None:
            return None

        cfg = getattr(self.model, "dlt_config", {}) or {}
        return_models = cfg.get("return_validated_models", False)

        if isinstance(item, list):
            validated_list = validate_and_filter_items(
                self.table_name, self.list_model, item, self.column_mode, self.data_mode
            )
            if return_models:
                return validated_list
            return [m.model_dump(by_alias=True) for m in validated_list]

        try:
            validated = validate_and_filter_item(
                self.table_name, self.model, item, self.column_mode, self.data_mode
            )
        except (ValidationError, DataValidationError):
            # pass through arrow, pandas and model items that pydantic cannot validate
            if get_data_item_format(item) != "object":
                return item
            raise
        if validated is None:
            return None

        if return_models:
            return validated

        return validated.model_dump(by_alias=True)

    def compute_table_schema(self, item: Any = None, meta: Any = None) -> Optional[TTableSchema]:
        """Computes authoritative table schema from the original Pydantic model.

        Only returns a schema when is_authoritative_model is set in DltConfig.
        """
        cfg = getattr(self.original_model, "dlt_config", None) or {}
        if not cfg.get("is_authoritative_model", False):
            return None

        from dlt.common.libs.pydantic import (
            pydantic_to_table_schema_columns,
            resolve_variant_model,
        )

        model = self.original_model
        if item is not None:
            model = resolve_variant_model(model, item, self._discriminator_map)  # type: ignore[assignment]
            if model is None:
                return None

        columns = pydantic_to_table_schema_columns(model)
        return {"name": self.table_name, "columns": columns}

    def __str__(self, *args: Any, **kwargs: Any) -> str:
        return f"PydanticValidator(model={self.model.__qualname__})"


def create_item_validator(
    columns: TTableHintTemplate[TAnySchemaColumns],
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
) -> Tuple[Optional[ValidateItem], TTableHintTemplate[TSchemaContract]]:
    """Creates item validator for a `columns` definition and a `schema_contract`

    Returns a tuple (validator, schema contract). If validator could not be created, returns None at first position.
    If schema_contract was not specified a default schema contract for given validator will be returned
    """
    if (
        PydanticBaseModel is not None
        and isinstance(columns, type)
        and issubclass(columns, PydanticBaseModel)
    ):
        assert not callable(
            schema_contract
        ), "schema_contract cannot be dynamic for Pydantic item validator"

        from dlt.common.libs.pydantic import (
            column_mode_to_extra,
            extra_to_column_mode,
            get_extra_from_model,
        )

        model_extra = get_extra_from_model(columns)
        model_column_mode = extra_to_column_mode(model_extra or "ignore")

        # warn when model explicitly sets extra and schema_contract contradicts it.
        # schema_contract can be a string (applies to all entities) or a dict
        if model_extra and schema_contract:
            if isinstance(schema_contract, str):
                explicit_columns = schema_contract
            elif isinstance(schema_contract, dict):
                explicit_columns = schema_contract.get("columns")
            if explicit_columns != model_column_mode:
                new_extra = column_mode_to_extra(explicit_columns)
                logger.warning(
                    f"Pydantic model {columns.__name__} has extra='{model_extra}' but the"
                    f" explicit schema_contract sets columns='{explicit_columns}'."
                    " The model's extra setting will be overridden to"
                    f" extra='{new_extra}' for data validation."
                    " Note that it is sufficient to just set the extra on the model "
                    "or columns on schema contract. dlt translates one setting into "
                    "another, refer to schema contract documentation."
                )

        # freeze the columns if we have a fully defined table and no other explicit contract
        expanded_schema_contract = Schema.expand_schema_contract_settings(
            schema_contract,
            # corresponds to default Pydantic behavior
            default={
                "tables": "evolve",
                "columns": model_column_mode,
                "data_type": "freeze",
            },
        )
        return (
            PydanticValidator(
                columns, expanded_schema_contract["columns"], expanded_schema_contract["data_type"]
            ),
            schema_contract or expanded_schema_contract,
        )
    return None, schema_contract



import dataclasses
from dlt import version
from typing import Final, Any, List, Dict, Optional, ClassVar

from dlt.common.configuration import configspec

from dlt.destinations.impl.mssql.configuration import (
    MsSqlCredentials,
    MsSqlClientConfiguration,
)

from dlt.destinations.impl.synapse.synapse_adapter import TTableIndexType


@configspec(init=False)
class SynapseCredentials(MsSqlCredentials):
    drivername: Final[str] = dataclasses.field(default="synapse", init=False, repr=False, compare=False)  # type: ignore

    # LongAsMax keyword got introduced in ODBC Driver 18 for SQL Server.
    SUPPORTED_DRIVERS: ClassVar[List[str]] = ["ODBC Driver 18 for SQL Server"]

    def get_odbc_dsn_dict(self) -> Dict[str, Any]:
        params = super().get_odbc_dsn_dict()
        # Long types (text, ntext, image) are not supported on Synapse.
        # Convert to max types using LongAsMax keyword.
        # https://stackoverflow.com/a/57926224
        params["LONGASMAX"] = "yes"
        return params


@configspec
class SynapseClientConfiguration(MsSqlClientConfiguration):
    destination_type: Final[str] = dataclasses.field(default="synapse", init=False, repr=False, compare=False)  # type: ignore
    credentials: SynapseCredentials = None

    # While Synapse uses CLUSTERED COLUMNSTORE INDEX tables by default, we use
    # HEAP tables (no indexing) by default. HEAP is a more robust choice, because
    # columnstore tables do not support varchar(max), nvarchar(max), and varbinary(max).
    # https://learn.microsoft.com/en-us/azure/synapse-analytics/sql-data-warehouse/sql-data-warehouse-tables-index
    default_table_index_type: Optional[TTableIndexType] = "heap"
    """
    Table index type that is used if no table index type is specified on the resource.
    This only affects data tables, dlt system tables ignore this setting and
    are always created as "heap" tables.
    """

    # Set to False by default because the PRIMARY KEY and UNIQUE constraints
    # are tricky in Synapse: they are NOT ENFORCED and can lead to innacurate
    # results if the user does not ensure all column values are unique.
    # https://learn.microsoft.com/en-us/azure/synapse-analytics/sql-data-warehouse/sql-data-warehouse-table-constraints
    create_indexes: bool = False
    """Whether `primary_key` and `unique` column hints are applied."""

    staging_use_msi: bool = False
    """Whether the managed identity of the Synapse workspace is used to authorize access to the staging Storage Account."""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "default_table_index_type",
        "create_indexes",
        "staging_use_msi",
    ]



from typing import (
    Tuple,
    cast,
    Any,
    Optional,
    Dict,
    Sequence,
    Mapping,
    List,
    NamedTuple,
)
from typing_extensions import Self

from dlt.common import logger
from dlt.common.schema.typing import (
    C_DLT_ID,
    TColumnProp,
    TFileFormat,
    TPartialTableSchema,
    TTableSchema,
    TTableSchemaColumns,
    TWriteDispositionConfig,
    TMergeDispositionDict,
    TScd2StrategyDict,
    TAnySchemaColumns,
    TTableFormat,
    TSchemaContract,
    DEFAULT_VALIDITY_COLUMN_NAMES,
    MERGE_STRATEGIES,
    TTableReferenceParam,
)
from dlt.common.exceptions import ValueErrorWithKnownValues
from dlt.common.typing import TTableNames, TypedDict, Unpack
from dlt.common.schema.utils import (
    DEFAULT_WRITE_DISPOSITION,
    is_nested_table,
    may_be_nested,
    merge_column,
    merge_columns,
    migrate_complex_types,
    new_column,
    new_table,
    remove_compound_props,
)
from dlt.common.typing import TAny, TDataItem, TColumnNames
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common.utils import clone_dict_nested
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.validation import validate_dict_ignoring_xkeys
from dlt.extract.exceptions import (
    DataItemRequiredForDynamicTableHints,
    InconsistentTableTemplate,
)
from dlt.extract.incremental import Incremental, TIncrementalConfig
from dlt.extract.items import TFunHintTemplate, TTableHintTemplate, TableNameMeta
from dlt.extract.items_transform import ValidateItem
from dlt.extract.utils import ensure_table_schema_columns, ensure_table_schema_columns_hint
from dlt.extract.validation import create_item_validator

import sqlglot


DLT_HINTS_METADATA_KEY = "dlt_resource_hints"


class TResourceNestedHints(TypedDict, total=False):
    # used to force a parent for rare cases where normalizer skips intermediate tables that
    # do not receive data
    parent_table_name: Optional[TTableHintTemplate[str]]
    write_disposition: Optional[TTableHintTemplate[TWriteDispositionConfig]]
    primary_key: Optional[TTableHintTemplate[TColumnNames]]
    columns: Optional[TTableHintTemplate[TAnySchemaColumns]]
    schema_contract: Optional[TTableHintTemplate[TSchemaContract]]
    table_format: Optional[TTableHintTemplate[TTableFormat]]
    file_format: Optional[TTableHintTemplate[TFileFormat]]
    merge_key: Optional[TTableHintTemplate[TColumnNames]]
    references: Optional[TTableHintTemplate[TTableReferenceParam]]


class TResourceHintsBase(TResourceNestedHints, total=False):
    table_name: Optional[TTableHintTemplate[str]]
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]]


class TResourceHints(TResourceHintsBase, total=False):
    # description: TTableHintTemplate[str]
    incremental: Optional[Incremental[Any]]
    validator: ValidateItem
    original_columns: Optional[TTableHintTemplate[TAnySchemaColumns]]
    additional_table_hints: Optional[Dict[str, TTableHintTemplate[Any]]]


class HintsMeta:
    __slots__ = ("hints", "create_table_variant")

    def __init__(
        self,
        hints: TResourceHints,
        create_table_variant: bool,
    ) -> None:
        self.hints = hints
        self.create_table_variant = create_table_variant


NATURAL_CALLABLES = ["incremental", "validator", "original_columns"]


def table_schema_to_hints(table: TTableSchema) -> TResourceHints:
    """Converts table schema into resource hints in place. Does not attempt to extract primary keys,
    additional table properties etc. into respective TResourceHints fields. Suitable only
    to be used for a result of utils.new_table.
    """
    template: TResourceHints = table  # type: ignore[assignment]
    # parent and name are different in table schema vs. resource hints so we need to rename
    template["table_name"] = table.pop("name")
    if "parent" in table:
        template["parent_table_name"] = table.pop("parent")
    # always remove resource
    table.pop("resource", None)
    return template


def make_nested_hints(**hints: Unpack[TResourceNestedHints]) -> TResourceNestedHints:
    return make_hints(**hints)


def make_hints(
    table_name: Optional[TTableHintTemplate[str]] = None,
    parent_table_name: Optional[TTableHintTemplate[str]] = None,
    write_disposition: Optional[TTableHintTemplate[TWriteDispositionConfig]] = None,
    columns: Optional[TTableHintTemplate[TAnySchemaColumns]] = None,
    primary_key: Optional[TTableHintTemplate[TColumnNames]] = None,
    merge_key: Optional[TTableHintTemplate[TColumnNames]] = None,
    schema_contract: Optional[TTableHintTemplate[TSchemaContract]] = None,
    table_format: Optional[TTableHintTemplate[TTableFormat]] = None,
    file_format: Optional[TTableHintTemplate[TFileFormat]] = None,
    additional_table_hints: Optional[Dict[str, TTableHintTemplate[Any]]] = None,
    references: Optional[TTableHintTemplate[TTableReferenceParam]] = None,
    incremental: Optional[TIncrementalConfig] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
) -> TResourceHints:
    """A convenience function to create resource hints. Accepts both static and dynamic hints based on data.

    This method accepts the same table hints arguments as `dlt.resource` decorator.
    """

    validator, schema_contract = create_item_validator(columns, schema_contract)
    # create a table schema template where hints can be functions taking TDataItem
    new_template = table_schema_to_hints(
        new_table(
            table_name,  # type: ignore
            parent_table_name,  # type: ignore
            write_disposition=write_disposition,  # type: ignore
            schema_contract=schema_contract,  # type: ignore
            table_format=table_format,  # type: ignore
            file_format=file_format,  # type: ignore
            references=references,  # type: ignore
        )
    )
    if not table_name:
        del new_template["table_name"]
    if not write_disposition and "write_disposition" in new_template:
        new_template.pop("write_disposition")
    # remember original columns and set template columns
    if columns is not None:
        new_template["original_columns"] = columns
        new_template["columns"] = ensure_table_schema_columns_hint(columns)

    if primary_key is not None:
        new_template["primary_key"] = primary_key
    if merge_key is not None:
        new_template["merge_key"] = merge_key
    if validator:
        new_template["validator"] = validator
    if nested_hints is not None:
        new_template["nested_hints"] = nested_hints
    if additional_table_hints is not None:
        new_template["additional_table_hints"] = additional_table_hints
    DltResourceHints.validate_dynamic_hints(new_template)
    if incremental is not None:  # TODO: Validate
        new_template["incremental"] = Incremental.ensure_instance(incremental)
    return new_template


class DltResourceHints:
    def __init__(self, table_schema_template: Optional[TResourceHints] = None):
        self.__qualname__ = self.__name__ = self.name
        self._table_name_hint_fun: Optional[TFunHintTemplate[str]] = None
        self._table_has_other_dynamic_hints: bool = False
        self._hints: Optional[TResourceHints] = None
        """Hints for the resource"""
        self._hints_variants: Dict[str, TResourceHints] = {}
        """Hints for tables emitted from resources"""
        if table_schema_template:
            self._set_hints(table_schema_template)

    @property
    def name(self) -> str:
        pass

    @property
    def table_name(self) -> TTableHintTemplate[str]:
        """Get table name to which resource loads data. May return a callable."""
        if self._table_name_hint_fun:
            return self._table_name_hint_fun
        # get table name or default name
        return self._hints.get("table_name") or self.name if self._hints else self.name

    @table_name.setter
    def table_name(self, value: TTableHintTemplate[str]) -> None:
        self.apply_hints(table_name=value)

    @property
    def has_dynamic_table_name(self) -> bool:
        """Tells the extractor wether computed table name may change based on invididual data items"""
        return self._table_name_hint_fun is not None

    @property
    def has_other_dynamic_hints(self) -> bool:
        """Tells the extractor wether computed hints may change based on invididual data items"""
        return self._table_has_other_dynamic_hints

    @property
    def write_disposition(self) -> TTableHintTemplate[TWriteDispositionConfig]:
        value = (self._hints or {}).get("write_disposition")
        return value if value is not None else DEFAULT_WRITE_DISPOSITION

    @write_disposition.setter
    def write_disposition(self, value: TTableHintTemplate[TWriteDispositionConfig]) -> None:
        self.apply_hints(write_disposition=value)

    @property
    def columns(self) -> TTableHintTemplate[TTableSchemaColumns]:
        """Gets columns' schema that can be modified in place"""
        return None if self._hints is None else self._hints.get("columns")  # type: ignore[return-value]

    @property
    def nested_hints(self) -> Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]]:
        return (
            None
            if self._hints is None or self._hints.get("nested_hints") is None
            else self._hints["nested_hints"]
        )

    @property
    def schema_contract(self) -> Optional[TTableHintTemplate[TSchemaContract]]:
        return None if self._hints is None else self._hints.get("schema_contract")

    @property
    def table_format(self) -> Optional[TTableHintTemplate[TTableFormat]]:
        return None if self._hints is None else self._hints.get("table_format")

    @property
    def parent_table_name(self) -> Optional[TTableHintTemplate[str]]:
        return None if self._hints is None else self._hints.get("parent_table_name")

    def compute_table_schema(self, item: TDataItem = None, meta: Any = None) -> TTableSchema:
        """Computes the table schema based on hints and column definitions passed during resource creation.
        `item` parameter is used to resolve table hints based on data.
        `meta` parameter is taken from Pipe and may further specify table name if variant is to be used
        """
        if isinstance(meta, TableNameMeta):
            # look for variant
            default_table_name = meta.table_name
            root_table_template = self._hints_variants.get(default_table_name, self._hints)
        else:
            default_table_name = self.name
            root_table_template = self._hints

        if not root_table_template:
            return new_table(default_table_name, resource=self.name)

        # resolve a copy of a held template
        root_table_template = self._clone_hints(root_table_template)
        if "table_name" not in root_table_template:
            root_table_template["table_name"] = default_table_name

        # if table template present and has dynamic hints, the data item must be provided.
        if self._table_name_hint_fun and item is None:
            raise DataItemRequiredForDynamicTableHints(self.name)

        # resolve
        resolved_template: TResourceHints = {
            k: self._resolve_hint(item, v)
            for k, v in root_table_template.items()
            if k not in NATURAL_CALLABLES
        }  # type: ignore

        if "incremental" in root_table_template:
            incremental = root_table_template["incremental"]
            if isinstance(incremental, Incremental) and incremental is not Incremental.EMPTY:
                resolved_template["incremental"] = incremental

        table_schema = self._hints_to_table_schema(resolved_template)
        if not is_nested_table(table_schema):
            table_schema["resource"] = self.name
        migrate_complex_types(table_schema, warn=True)
        validate_dict_ignoring_xkeys(
            spec=TTableSchema,
            doc=table_schema,
            path=f"new_table/{self.name}",
        )

        return table_schema

    def compute_nested_table_schemas(
        self,
        root_table_name: str,
        naming: NamingConvention,
        item: TDataItem = None,
        meta: Any = None,
    ) -> List[TTableSchema]:
        """Compute the table schema based on the current and all nested hints.
        Nested hints are resolved recursively.
        """
        # allow for nested hints to be a part of a table variant
        if isinstance(meta, TableNameMeta):
            root_table_template = self._hints_variants.get(root_table_name, self._hints)
        else:
            root_table_template = self._hints
        # no nested hints if no table template
        if root_table_template is None:
            return []

        nested_hints = root_table_template.get("nested_hints")
        if not nested_hints:
            return []
        # resolve dynamic hint with actual data item
        nested_hints = self._resolve_hint(item, nested_hints)

        nested_table_schemas = []
        # make sure shorter paths go first
        sorted_items = sorted(
            nested_hints.items(), key=lambda item: len(item[0]) if isinstance(item[0], tuple) else 1
        )
        for sub_path, hints in sorted_items:
            # both str and sequence is supported
            if isinstance(sub_path, str):
                full_path: Tuple[str, ...] = (root_table_name, sub_path)
            else:
                full_path = (root_table_name, *sub_path)
            table_name = naming.shorten_fragments(*full_path)
            nested_table_template = self._clone_hints(hints)

            resolved_template: TResourceHints = {
                k: self._resolve_hint(item, v)
                for k, v in nested_table_template.items()
                if k not in NATURAL_CALLABLES
            }  # type: ignore

            nested_table_schema = self._hints_to_table_schema(resolved_template)
            # table_name is not a part of TResourceNestedHints but in the future we may allow those tables
            # to be renamed. that will require a bigger refactor in the relational normalizer
            nested_table_schema["name"] = table_name
            migrate_complex_types(nested_table_schema, warn=True)

            # in very rare cases parent may be explicitly defined for a nested table
            if "parent" not in nested_table_schema and may_be_nested(nested_table_schema):
                # we add default parent linking if table is eligible
                nested_table_schema["parent"] = naming.shorten_fragments(*full_path[:-1])

            if not is_nested_table(nested_table_schema):
                nested_table_schema["resource"] = self.name
            elif nested_write_disposition := nested_table_schema.get("write_disposition"):
                write_disposition = root_table_template.get("write_disposition")
                # TODO: write_disposition may also be a callable, in that case generate warning as well
                if nested_write_disposition != write_disposition:
                    logger.warning(
                        f"You defined '{nested_write_disposition}' write disposition on a nested"
                        f" table {table_name} while root table {root_table_name} has"
                        f" {write_disposition} write disposition. This may lead to data"
                        " inconsistency across root and nested tables. Consider setting primary or"
                        " merge key on a nested table to break nesting chain and convert it into a"
                        "  top level table."
                    )

            validate_dict_ignoring_xkeys(
                spec=TTableSchema,
                doc=nested_table_schema,
                path=f"new_table/{table_name}",
            )
            nested_table_schemas.append(nested_table_schema)
        # TODO: restore this implicit table creation but initially I'd push it to
        # the user to create intermediate tables
        # NOTE we insert missing parents at the beginning to ensure that `schema.update_table()` is later called
        # in an an order that respect parent-child relationships
        # for table_name in tables_to_create:
        #     # TODO need to clean up the path from args -> hints -> schema
        #     placeholder_table_template = new_table(table_name=table_name)
        #     placeholder_table_schema = self._create_table_schema(placeholder_table_template, self.name)
        #     nested_table_schemas.insert(0, placeholder_table_schema)

        return nested_table_schemas

    def apply_hints(
        self,
        table_name: Optional[TTableHintTemplate[str]] = None,
        parent_table_name: Optional[TTableHintTemplate[str]] = None,
        write_disposition: Optional[TTableHintTemplate[TWriteDispositionConfig]] = None,
        columns: Optional[TTableHintTemplate[TAnySchemaColumns]] = None,
        primary_key: Optional[TTableHintTemplate[TColumnNames]] = None,
        merge_key: Optional[TTableHintTemplate[TColumnNames]] = None,
        incremental: Optional[TIncrementalConfig] = None,
        schema_contract: Optional[TTableHintTemplate[TSchemaContract]] = None,
        additional_table_hints: Optional[Dict[str, TTableHintTemplate[Any]]] = None,
        table_format: Optional[TTableHintTemplate[TTableFormat]] = None,
        file_format: Optional[TTableHintTemplate[TFileFormat]] = None,
        references: Optional[TTableHintTemplate[TTableReferenceParam]] = None,
        create_table_variant: bool = False,
        nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    ) -> Self:
        """Creates or modifies existing table schema by setting provided hints. Accepts both static and dynamic hints based on data.

        If `create_table_variant` is specified, the `table_name` must be a string and hints will be used to create a separate set of hints
        for a particular `table_name`. Such hints may be retrieved via compute_table_schema(meta=TableNameMeta(table_name)).
        Table variant hints may not contain dynamic hints.

        This method accepts the same table hints arguments as `dlt.resource` decorator with the following additions.
        Skip the argument or pass None to leave the existing hint.
        Pass empty value (for a particular type i.e. "" for a string) to remove a hint.

        Args:
            table_name (TTableHintTemplate[str]): name of the table which resource will generate

            parent_table_parent (str, optional): A name of parent table if you want the resource to generate nested table. Please note that if you use merge, you must define `root_key` columns explicitly

            incremental (Incremental, optional): Enables the incremental loading for a resource.

        Please note that for efficient incremental loading, the resource must be aware of the Incremental by accepting it as one if its arguments and then using are to skip already loaded data.
        In non-aware resources, `dlt` will filter out the loaded values, however, the resource will yield all the values again.

        Returns: self for chaining
        """
        if create_table_variant:
            if not isinstance(table_name, str):
                raise ValueError(
                    "Please provide `str` table name if you want to create a table variant of hints"
                )
            # select hints variant
            t = self._hints_variants.get(table_name, None)
            if t is None:
                # use resource hints as starting point
                if self._hints:
                    t = self._clone_hints(self._hints)
                    # but remove callables
                    t = {n: h for n, h in t.items() if not callable(h)}  # type: ignore[assignment]
        else:
            t = self._hints

        if t is None:
            # if there is no template yet, create and set a new one.
            default_wd = None if parent_table_name else DEFAULT_WRITE_DISPOSITION
            t = make_hints(
                table_name=table_name,
                parent_table_name=parent_table_name,
                write_disposition=write_disposition or default_wd,
                columns=columns,
                primary_key=primary_key,
                merge_key=merge_key,
                schema_contract=schema_contract,
                table_format=table_format,
                file_format=file_format,
                references=references,
                nested_hints=nested_hints,
            )
        else:
            t = self._clone_hints(t)
            if table_name is not None:
                if table_name:
                    t["table_name"] = table_name
                else:
                    t.pop("table_name", None)
            if parent_table_name is not None:
                if parent_table_name:
                    t["parent_table_name"] = parent_table_name
                else:
                    t.pop("parent_table_name", None)
            if write_disposition:
                t["write_disposition"] = write_disposition
            if columns is not None:
                # keep original columns: i.e. in case it is a Pydantic model.
                t["original_columns"] = columns
                # if callable then override existing
                if callable(columns) or callable(t["columns"]):
                    t["columns"] = ensure_table_schema_columns_hint(columns)
                elif columns:
                    # normalize columns
                    columns = ensure_table_schema_columns(columns)
                    # this updates all columns with defaults
                    assert isinstance(t["columns"], dict)
                    t["columns"] = merge_columns(t["columns"], columns)
                else:
                    # set to empty columns
                    t["columns"] = ensure_table_schema_columns(columns)
            if primary_key is not None:
                t["primary_key"] = primary_key
            if merge_key is not None:
                t["merge_key"] = merge_key
            if schema_contract is not None:
                if schema_contract:
                    t["schema_contract"] = schema_contract
                else:
                    t.pop("schema_contract", None)
            if additional_table_hints is not None:
                if additional_table_hints:
                    if t.get("additional_table_hints") is not None:
                        for k, v in additional_table_hints.items():
                            t["additional_table_hints"][k] = v
                    else:
                        t["additional_table_hints"] = additional_table_hints
                else:
                    t.pop("additional_table_hints", None)

            # recreate validator if column definition or contract changed
            if schema_contract is not None or columns is not None:
                t["validator"], schema_contract = create_item_validator(
                    t.get("original_columns"), t.get("schema_contract")
                )
            if schema_contract is not None:
                t["schema_contract"] = schema_contract
            if table_format is not None:
                if table_format:
                    t["table_format"] = table_format
                else:
                    t.pop("table_format", None)
            if file_format is not None:
                if file_format:
                    t["file_format"] = file_format
                else:
                    t.pop("file_format", None)
            if references is not None:
                if callable(references) or callable(t.get("references")):
                    t["references"] = references
                else:
                    # Replace existing refs for same table
                    new_references = t.get("references") or []
                    ref_dict = {r["referenced_table"]: r for r in new_references}  # type: ignore[union-attr]
                    for ref in references:
                        ref_dict[ref["referenced_table"]] = ref
                    t["references"] = list(ref_dict.values())
            # NOTE: here we just replace nested hints fully (or drop them). what we could do instead
            #   is to re-use the code above (ie special handling of references or columns) for each nested hint
            #   for now we are good
            if nested_hints is not None:
                if nested_hints:
                    t["nested_hints"] = nested_hints
                else:
                    t.pop("nested_hints", None)

        # set properties that can't be passed to make_hints
        if incremental is not None:
            t["incremental"] = Incremental.ensure_instance(incremental)

        self._set_hints(t, create_table_variant)
        return self

    def _set_hints(
        self, hints_template: TResourceHints, create_table_variant: bool = False
    ) -> None:
        DltResourceHints.validate_dynamic_hints(hints_template)
        DltResourceHints.validate_write_disposition_hint(hints_template)
        DltResourceHints.validate_reference_hint(hints_template)
        if create_table_variant:
            # for table variants, table name must be a str
            table_name: str = hints_template["table_name"]  # type: ignore[assignment]
            # incremental cannot be specified in variant
            if hints_template.get("incremental"):
                raise InconsistentTableTemplate(
                    f"You can specify `incremental` only for the resource `{self.name}` hints, not"
                    f" in table `{table_name}` variant-"
                )
            if hints_template.get("validator"):
                logger.warning(
                    f"A data item validator was created from column schema in `{self.name}` for a"
                    f" table `{table_name}` variant. Currently such validator is ignored."
                )
            # dynamic hints will be ignored
            for name, hint in hints_template.items():
                if callable(hint) and name not in NATURAL_CALLABLES:
                    raise InconsistentTableTemplate(
                        f"Table `{table_name}` variant hint is resource `{self.name}` can't have"
                        f" dynamic hint but `{name}` does."
                    )
            self._hints_variants[table_name] = hints_template
        else:
            # if "name" is callable in the template, then the table schema requires data item to be inferred.
            name_hint = hints_template.get("table_name")
            self._table_name_hint_fun = name_hint if callable(name_hint) else None
            # check if any other hints in the table template should be inferred from data.
            self._table_has_other_dynamic_hints = any(
                callable(v) for k, v in hints_template.items() if k != "table_name"
            )
            self._hints = hints_template

    def merge_hints(
        self, hints_template: TResourceHints, create_table_variant: bool = False
    ) -> None:
        self.apply_hints(
            table_name=hints_template.get("table_name"),
            parent_table_name=hints_template.get("parent_table_name"),
            write_disposition=hints_template.get("write_disposition"),
            columns=hints_template.get("original_columns"),
            primary_key=hints_template.get("primary_key"),
            merge_key=hints_template.get("merge_key"),
            incremental=hints_template.get("incremental"),
            schema_contract=hints_template.get("schema_contract"),
            table_format=hints_template.get("table_format"),
            file_format=hints_template.get("file_format"),
            additional_table_hints=hints_template.get("additional_table_hints"),
            references=hints_template.get("references"),
            create_table_variant=create_table_variant,
            nested_hints=hints_template.get("nested_hints"),
        )

    @staticmethod
    def _clone_hints(hints_template: TAny) -> TAny:
        if hints_template is None:
            return None
        # creates a deep copy of dict structure without actually copying the objects
        return clone_dict_nested(hints_template)

    @staticmethod
    def _resolve_hint(item: TDataItem, hint: TTableHintTemplate[TAny]) -> TAny:
        """Calls each dynamic hint passing a data item"""
        return hint(item) if callable(hint) else hint  # type: ignore[no-any-return]

    @staticmethod
    def _merge_key(hint: TColumnProp, keys: TColumnNames, partial: TPartialTableSchema) -> None:
        remove_compound_props(partial["columns"], {hint})
        if not keys:
            return
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if key in partial["columns"]:
                # set nullable to False if not set
                nullable = partial["columns"][key].get("nullable", False)
                merge_column(partial["columns"][key], {hint: True, "nullable": nullable})  # type: ignore
            else:
                partial["columns"][key] = new_column(key, nullable=False)
                partial["columns"][key][hint] = True

    @staticmethod
    def _merge_keys(dict_: TResourceHints) -> None:
        """Applies primary_key and merge_key hints to columns. Table-level hints override column-level settings."""

        if "primary_key" in dict_:
            DltResourceHints._merge_key("primary_key", dict_.pop("primary_key"), dict_)  # type: ignore
        if "merge_key" in dict_:
            DltResourceHints._merge_key("merge_key", dict_.pop("merge_key"), dict_)  # type: ignore

    @staticmethod
    def _merge_write_disposition_dict(dict_: Dict[str, Any]) -> None:
        """Merges write disposition dictionary into write disposition shorthand and x-hints in place."""

        write_disposition = dict_["write_disposition"]["disposition"]
        if write_disposition == "merge":
            DltResourceHints._merge_merge_disposition_dict(dict_)
        # reduce merge disposition from dict to shorthand
        dict_["write_disposition"] = write_disposition

    @staticmethod
    def _merge_merge_disposition_dict(dict_: Dict[str, Any]) -> None:
        """Merges merge disposition dict into x-hints in place."""

        md_dict: TMergeDispositionDict = dict_.pop("write_disposition")
        if merge_strategy := md_dict.get("strategy"):
            dict_["x-merge-strategy"] = merge_strategy

        if deduplicated := md_dict.get("deduplicated"):
            dict_["x-stage-data-deduplicated"] = deduplicated

        if merge_strategy == "scd2":
            md_dict = cast(TScd2StrategyDict, md_dict)
            if "boundary_timestamp" in md_dict:
                dict_["x-boundary-timestamp"] = md_dict["boundary_timestamp"]
            if md_dict.get("validity_column_names") is None:
                from_, to = DEFAULT_VALIDITY_COLUMN_NAMES
            else:
                from_, to = md_dict["validity_column_names"]
            dict_["columns"][from_] = {
                "name": from_,
                "data_type": "timestamp",
                "nullable": True,  # validity columns are empty when first loaded into staging table
                "x-valid-from": True,
            }
            dict_["columns"][to] = {
                "name": to,
                "data_type": "timestamp",
                "nullable": True,
                "x-valid-to": True,
                "x-active-record-timestamp": md_dict.get("active_record_timestamp"),
            }
            # unique constraint is dropped for C_DLT_ID when used to store
            # SCD2 row hash (only applies to root table)
            hash_ = md_dict.get("row_version_column_name", C_DLT_ID)
            dict_["columns"][hash_] = {
                "name": hash_,
                "nullable": False,
                "x-row-version": True,
                # duplicate value in row hash column is possible in case
                # of insert-delete-reinsert pattern
                "unique": False,
                "row_key": False,
            }

    @staticmethod
    def _merge_incremental_column_hint(dict_: Dict[str, Any]) -> None:
        incremental = dict_.pop("incremental")
        if incremental is None:
            return
        col_name = incremental.get_cursor_column_name()
        if not col_name:
            # cursor cannot resolve to a single column, no hint added
            return
        incremental_col = dict_["columns"].get(col_name)
        if not incremental_col:
            incremental_col = {"name": col_name}

        incremental_col["incremental"] = True
        dict_["columns"][col_name] = incremental_col

    @staticmethod
    def _hints_to_table_schema(resolved_hints: TResourceHints) -> TTableSchema:
        """Creates table schema from resource hints in place. Resource hints must be resolved
        (cannot contain callables).
        """
        resolved_hints.pop("nested_hints", None)
        resolved_hints["columns"] = resolved_hints.get("columns", {})
        DltResourceHints._merge_keys(resolved_hints)
        if "write_disposition" in resolved_hints:
            if isinstance(resolved_hints["write_disposition"], str):
                resolved_hints["write_disposition"] = {
                    "disposition": resolved_hints["write_disposition"]
                }  # wrap in dict
            DltResourceHints._merge_write_disposition_dict(resolved_hints)  # type: ignore[arg-type]
        if "incremental" in resolved_hints:
            DltResourceHints._merge_incremental_column_hint(resolved_hints)  # type: ignore[arg-type]
        table = cast(TTableSchema, resolved_hints)
        # rename two props that differ
        if "table_name" in resolved_hints:
            table["name"] = resolved_hints.pop("table_name")  # type: ignore[typeddict-item]
        if "parent_table_name" in resolved_hints:
            table["parent"] = resolved_hints.pop("parent_table_name")  # type: ignore[typeddict-item]
        # apply table hints
        if additional_table_hints := resolved_hints.get("additional_table_hints"):
            for k, v in additional_table_hints.items():
                table[k] = v  # type: ignore[literal-required]
        resolved_hints.pop("additional_table_hints", None)
        return table

    @staticmethod
    def validate_dynamic_hints(template: TResourceHints) -> None:
        table_name = template.get("table_name")
        # if any of the hints is a function, then name must be as well.
        if any(
            callable(v) for k, v in template.items() if k not in ["table_name", *NATURAL_CALLABLES]
        ) and not callable(table_name):
            raise InconsistentTableTemplate(
                f"Table name `{table_name}` must be a function if any other table hint is a"
                " function"
            )

    @staticmethod
    def validate_write_disposition_hint(template: TResourceHints) -> None:
        wd = template.get("write_disposition")
        if isinstance(wd, dict) and wd["disposition"] == "merge":
            wd = cast(TMergeDispositionDict, wd)
            if "strategy" in wd and wd["strategy"] not in MERGE_STRATEGIES:
                raise ValueErrorWithKnownValues(
                    "write_disposition['strategy']", wd["strategy"], MERGE_STRATEGIES
                )

            if wd.get("strategy") == "scd2":
                wd = cast(TScd2StrategyDict, wd)
                for ts in ("active_record_timestamp", "boundary_timestamp"):
                    if (
                        ts == "active_record_timestamp"
                        and wd.get("active_record_timestamp") is None
                    ):
                        continue  # None is allowed for active_record_timestamp
                    if ts in wd:
                        if wd[ts] is None:  # type: ignore[literal-required]
                            continue
                        try:
                            ensure_pendulum_datetime_utc(wd[ts])  # type: ignore[literal-required]
                        except Exception:
                            raise ValueError(
                                f"could not parse `{ts}` value `{wd[ts]}`"  # type: ignore[literal-required]
                            )

    @staticmethod
    def validate_reference_hint(template: TResourceHints) -> None:
        ref = template.get("reference")
        if ref is None:
            return
        if not isinstance(ref, Sequence):
            raise ValueError("Reference hint must be a sequence of table references.")
        for r in ref:
            if not isinstance(r, Mapping):
                raise ValueError("Table reference must be a dictionary.")
            columns = r.get("columns")
            referenced_columns = r.get("referenced_columns")
            table = r.get("referenced_table")
            if not table:
                raise ValueError("Referenced table must be specified.")
            if not columns or not referenced_columns:
                raise ValueError("Both columns and referenced_columns must be specified.")
            if len(columns) != len(referenced_columns):
                raise ValueError("Columns and referenced_columns must have the same length.")



from typing import Any, Sequence

from dlt.common.reflection.ref import ImportTrace


class ReferenceImportError:
    """A mixing for your exception to store and pretty print ref import traces."""

    def __init__(self, *args: Any, **kwargs: Any):
        self.traces: Sequence[ImportTrace] = kwargs.get("traces")

    def __str__(self) -> str:
        if not self.traces:
            return "No import traces were corrected"
        msg = "Modules and attributes were tried in the following order and failed to import:\n"
        for trace in self.traces:
            msg += f"\tmod:`{trace.module}` attr: `{trace.attr_name}` failed due to: {trace.reason}"
            if trace.exc:
                msg += f" and causing exception: {trace.exc}"
            msg += "\n"

        return msg



from typing import Any, Optional, Literal, Dict, Union, Sequence

from dateutil import parser

from dlt.common.destination import PreparedTableSchema
from dlt.common.pendulum import timezone
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.typing import TColumnNames
from dlt.destinations.utils import get_resource_for_adapter
from dlt.extract import DltResource
from dlt.extract.items import TTableHintTemplate


PARTITION_HINT: Literal["x-bigquery-partition"] = "x-bigquery-partition"
CLUSTER_HINT: Literal["x-bigquery-cluster"] = "x-bigquery-cluster"
ROUND_HALF_AWAY_FROM_ZERO_HINT: Literal["x-bigquery-round-half-away-from-zero"] = (
    "x-bigquery-round-half-away-from-zero"
)
ROUND_HALF_EVEN_HINT: Literal["x-bigquery-round-half-even"] = "x-bigquery-round-half-even"
TABLE_EXPIRATION_HINT: Literal["x-bigquery-table-expiration"] = "x-bigquery-table-expiration"
TABLE_DESCRIPTION_HINT: Literal["x-bigquery-table-description"] = "x-bigquery-table-description"
AUTODETECT_SCHEMA_HINT: Literal["x-bigquery-autodetect-schema"] = "x-bigquery-autodetect-schema"
PARTITION_EXPIRATION_DAYS_HINT: Literal["x-bigquery-partition-expiration-days"] = (
    "x-bigquery-partition-expiration-days"
)
CLUSTER_COLUMNS_HINT: Literal["x-bigquery-cluster-columns"] = "x-bigquery-cluster-columns"


class PartitionTransformation:
    template: str
    """Template string of the transformation including column name placeholder. E.g. `RANGE_BUCKET({column_name}, GENERATE_ARRAY(0, 1000000, 10000))`"""
    column_name: str
    """Column name to apply the transformation to"""

    def __init__(self, template: str, column_name: str) -> None:
        self.template = template
        self.column_name = column_name


class bigquery_partition:
    """Helper class to generate BigQuery partition transformations."""

    @staticmethod
    def range_bucket(
        column_name: str,
        start: int,
        end: int,
        interval: int = 1,
    ) -> PartitionTransformation:
        """Partition by an integer column with the specified range, where:

        Args:
            column_name: The column to partition by
            start: The start of range partitioning (inclusive)
            end: The end of the range partitioning (exclusive)
            interval: The width of each range within the partition (default: 1)

        Returns:
            A PartitionTransformation object for integer range partitioning
        """

        template = f"RANGE_BUCKET({{column_name}}, GENERATE_ARRAY({start}, {end}, {interval}))"

        return PartitionTransformation(template, column_name)


def _set_column_hint(
    column_hints: TTableSchemaColumns, column_name: str, hint_key: str, value: Any = True
) -> None:
    """Add a hint to a column, initializing the column dict if needed."""
    column_hints.setdefault(column_name, {"name": column_name})[hint_key] = value  # type: ignore[literal-required]


def bigquery_adapter(
    data: Any,
    partition: Union[TColumnNames, PartitionTransformation] = None,
    cluster: TColumnNames = None,
    round_half_away_from_zero: TColumnNames = None,
    round_half_even: TColumnNames = None,
    table_description: Optional[str] = None,
    table_expiration_datetime: Optional[str] = None,
    insert_api: Optional[Literal["streaming", "default"]] = None,
    autodetect_schema: Optional[bool] = None,
    partition_expiration_days: Optional[int] = None,
) -> DltResource:
    """
    Prepares data for loading into BigQuery.

    This function takes data, which can be raw or already wrapped in a DltResource object,
    and prepares it for BigQuery by optionally specifying partitioning, clustering, table description and
    table expiration settings.

    Args:
        data (Any): The data to be transformed.
            This can be raw data or an instance of DltResource.
            If raw data is provided, the function will wrap it into a `DltResource` object.
        partition (Union[TColumnNames, PartitionTransformation], optional): The column to partition the BigQuery table by.
            This can be a string representing a single column name for simple partitioning,
            or a PartitionTransformation object for advanced partitioning.
            Use the bigquery_partition helper class to create transformation objects.
        cluster (TColumnNames, optional): A column name or list of column names to cluster the BigQuery table by.
        round_half_away_from_zero (TColumnNames, optional): Determines how values in the column are rounded when written to the table.
            This mode rounds halfway cases away from zero.
            The columns specified must be mutually exclusive from `round_half_even`.
            See https://cloud.google.com/bigquery/docs/schemas#rounding_mode for more information.
        round_half_even (TColumnNames, optional): Determines how values in the column are rounded when written to the table.
            This mode rounds halfway cases towards the nearest even digit.
            The columns specified must be mutually exclusive from `round_half_away_from_zero`.
            See https://cloud.google.com/bigquery/docs/schemas#rounding_mode for more information.
        table_description (str, optional): A description for the BigQuery table.
        table_expiration_datetime (str, optional): String representing the datetime when the BigQuery table expires.
            This is always interpreted as UTC, BigQuery's default.
        insert_api (Optional[Literal["streaming", "default"]]): The API to use for inserting data into BigQuery.
            If "default" is chosen, the original SQL query mechanism is used.
            If "streaming" is chosen, the streaming API (https://cloud.google.com/bigquery/docs/streaming-data-into-bigquery)
            is used.
            NOTE: due to BigQuery features, streaming insert is only available for `append` write_disposition.
        autodetect_schema (bool, optional): If set to True, BigQuery schema autodetection will be used to create data tables. This
            allows to create structured types from nested data.
        partition_expiration_days (int, optional): For date/time based partitions it tells when partition is expired and removed.
            Partitions are expired based on a partitioned column value. (https://cloud.google.com/bigquery/docs/managing-partitioned-tables#partition-expiration)

    Returns:
        A `DltResource` object that is ready to be loaded into BigQuery.

    Raises:
        ValueError: If any hint is invalid or none are specified.

    Examples:
        >>> data = [{"name": "Marcel", "description": "Raccoon Engineer", "date_hired": 1700784000}]
        >>> bigquery_adapter(data, partition="date_hired", table_expiration_datetime="2024-01-30", table_description="Employee Data")
        [DltResource with hints applied]
    """
    resource = get_resource_for_adapter(data)

    additional_table_hints: Dict[str, TTableHintTemplate[Any]] = {}
    column_hints: TTableSchemaColumns = {}

    if partition:
        if not (isinstance(partition, str) or isinstance(partition, PartitionTransformation)):
            raise ValueError(
                "`partition` must be a single column name as a `str` or a"
                " `PartitionTransformation`."
            )

        # Can only have one partition column.
        for column in resource.columns.values():  # type: ignore[union-attr]
            column.pop(PARTITION_HINT, None)  # type: ignore[typeddict-item]

        if isinstance(partition, str):
            _set_column_hint(column_hints, partition, PARTITION_HINT)

        if isinstance(partition, PartitionTransformation):
            partition_hint: Dict[str, str] = {}
            partition_hint[partition.column_name] = partition.template
            additional_table_hints[PARTITION_HINT] = partition_hint

    if cluster:
        if isinstance(cluster, str):
            cluster = [cluster]
        if not isinstance(cluster, list):
            raise ValueError(
                "`cluster` must be a list of column names or a single column name as a string."
            )
        for column_name in cluster:
            _set_column_hint(column_hints, column_name, CLUSTER_HINT)
        additional_table_hints[CLUSTER_COLUMNS_HINT] = cluster

    # Implementing rounding logic flags
    if round_half_away_from_zero:
        if isinstance(round_half_away_from_zero, str):
            round_half_away_from_zero = [round_half_away_from_zero]
        if not isinstance(round_half_away_from_zero, list):
            raise ValueError(
                "`round_half_away_from_zero` must be a list of column names or a single column"
                " name."
            )
        for column_name in round_half_away_from_zero:
            _set_column_hint(column_hints, column_name, ROUND_HALF_AWAY_FROM_ZERO_HINT)

    if round_half_even:
        if isinstance(round_half_even, str):
            round_half_even = [round_half_even]
        if not isinstance(round_half_even, list):
            raise ValueError(
                "`round_half_even` must be a list of column names or a single column name."
            )
        for column_name in round_half_even:
            _set_column_hint(column_hints, column_name, ROUND_HALF_EVEN_HINT)

    if round_half_away_from_zero and round_half_even:
        if intersection_columns := set(round_half_away_from_zero).intersection(
            set(round_half_even)
        ):
            raise ValueError(
                f"Columns `{intersection_columns}` are present in both `round_half_away_from_zero`"
                " and `round_half_even` which is not allowed. They must be mutually exclusive."
            )

    if table_description:
        if not isinstance(table_description, str):
            raise ValueError(
                "`table_description` must be string representing BigQuery table description."
            )
        additional_table_hints[TABLE_DESCRIPTION_HINT] = table_description

    if autodetect_schema:
        additional_table_hints[AUTODETECT_SCHEMA_HINT] = autodetect_schema

    if table_expiration_datetime:
        if not isinstance(table_expiration_datetime, str):
            raise ValueError(
                "`table_expiration_datetime` must be string representing the datetime when the"
                " BigQuery table will be deleted."
            )
        try:
            parsed_table_expiration_datetime = parser.parse(table_expiration_datetime).replace(
                tzinfo=timezone.utc
            )
            additional_table_hints[TABLE_EXPIRATION_HINT] = parsed_table_expiration_datetime
        except ValueError as e:
            raise ValueError(
                f"`table_expiration_datetime={table_expiration_datetime}` could not be parsed!"
            ) from e

    if partition_expiration_days is not None:
        assert isinstance(
            partition_expiration_days, int
        ), "partition_expiration_days must be an integer (days)"
        additional_table_hints[PARTITION_EXPIRATION_DAYS_HINT] = partition_expiration_days

    if insert_api is not None:
        if insert_api == "streaming" and data.write_disposition != "append":
            raise ValueError(
                "BigQuery streaming insert only accepts `write_disposition='append'`. "
                f"Received `write_disposition={data.write_disposition}`."
            )
        additional_table_hints["x-insert-api"] = insert_api

    if column_hints or additional_table_hints:
        resource.apply_hints(columns=column_hints, additional_table_hints=additional_table_hints)
    else:
        raise ValueError(
            "AT LEAST one of `partition`, `cluster`, `round_half_away_from_zero`,"
            " `round_half_even`, `table_description` or `table_expiration_datetime` must be"
            " specified."
        )
    return resource


def should_autodetect_schema(table: PreparedTableSchema) -> bool:
    """Tells if schema should be auto detected for a given prepared `table`"""
    return table.get(AUTODETECT_SCHEMA_HINT, False)  # type: ignore[return-value]



from abc import ABC, abstractmethod
import os
import tempfile  # noqa: 251
from typing import Dict, Iterable, List

from dlt.common import pendulum
from dlt.common.json import json
from dlt.common.destination.client import (
    HasFollowupJobs,
    TLoadJobState,
    RunnableLoadJob,
    FollowupJobRequest,
    LoadJob,
)
from dlt.common.storages.load_package import commit_load_package_state
from dlt.common.storages import FileStorage
from dlt.common.typing import TDataItems
from dlt.common.storages.load_storage import ParsedLoadJobFileName

from dlt.destinations.impl.destination.configuration import (
    CustomDestinationClientConfiguration,
    TDestinationCallable,
)


class FinalizedLoadJob(LoadJob):
    """
    Special Load Job that should never get started and just indicates a job being in a final state.
    May also be used to indicate that nothing needs to be done.
    """

    def __init__(
        self,
        file_path: str,
        /,
        *,
        started_at: pendulum.DateTime = None,
        finished_at: pendulum.DateTime = None,
        status: TLoadJobState = "completed",
        failed_message: str = None,
        exception: BaseException = None,
    ) -> None:
        super().__init__(file_path)
        self._status = status
        self._failed_message = failed_message
        self._exception = exception
        self._started_at = started_at or pendulum.now()
        self._finished_at = finished_at or (
            pendulum.now() if self._status in ("completed", "failed") else None
        )
        assert self._status in ("completed", "failed", "retry")

    @classmethod
    def from_file_path(
        cls,
        file_path: str,
        /,
        *,
        started_at: pendulum.DateTime = None,
        finished_at: pendulum.DateTime = None,
        status: TLoadJobState = "completed",
        message: str = None,
        exception: BaseException = None,
    ) -> "FinalizedLoadJob":
        return cls(
            file_path,
            started_at=started_at,
            finished_at=finished_at,
            status=status,
            failed_message=message,
            exception=exception,
        )

    def state(self) -> TLoadJobState:
        return self._status

    def failed_message(self) -> str:
        return self._failed_message

    def exception(self) -> BaseException:
        return self._exception


class FinalizedLoadJobWithFollowupJobs(FinalizedLoadJob, HasFollowupJobs):
    pass


class FollowupJobRequestImpl(FollowupJobRequest):
    """
    Class to create a new loadjob, not stateful and not runnable
    """

    def __init__(self, file_name: str) -> None:
        self._file_path = os.path.join(tempfile.gettempdir(), file_name)
        self._parsed_file_name = ParsedLoadJobFileName.parse(file_name)
        # we only accept jobs that we can schedule as new or mark as failed.

    def _save_text_file(self, data: str) -> None:
        with open(self._file_path, "w", encoding="utf-8") as f:
            f.write(data)

    def new_file_path(self) -> str:
        """Path to a newly created temporary job file"""
        return self._file_path

    def job_id(self) -> str:
        """The job id that is derived from the file name and does not changes during job lifecycle"""
        return self._parsed_file_name.job_id()


class ReferenceFollowupJobRequest(FollowupJobRequestImpl):
    def __init__(self, original_file_name: str, remote_paths: List[str]) -> None:
        job_info = ParsedLoadJobFileName.parse(original_file_name)
        file_name = job_info.to_reference_file_name()
        self._remote_paths = remote_paths
        super().__init__(file_name)
        self._save_text_file("\n".join(remote_paths))

    @staticmethod
    def is_reference_job(file_path: str) -> bool:
        return os.path.splitext(file_path)[1][1:] == "reference"

    @staticmethod
    def resolve_references(file_path: str) -> List[str]:
        with open(file_path, "r+", encoding="utf-8") as f:
            # Reading from a file
            return f.read().split("\n")

    @staticmethod
    def resolve_reference(file_path: str) -> str:
        refs = ReferenceFollowupJobRequest.resolve_references(file_path)
        assert len(refs) == 1
        return refs[0]


class DestinationLoadJob(RunnableLoadJob, ABC):
    def __init__(
        self,
        file_path: str,
        config: CustomDestinationClientConfiguration,
        destination_state: Dict[str, int],
        destination_callable: TDestinationCallable,
        skipped_columns: List[str],
        callable_requires_job_client_args: bool = False,
    ) -> None:
        super().__init__(file_path)
        self._config = config
        self._callable = destination_callable
        self._storage_id = f"{self._parsed_file_name.table_name}.{self._parsed_file_name.file_id}"
        self._skipped_columns = skipped_columns
        self._destination_state = destination_state
        self._callable_requires_job_client_args = callable_requires_job_client_args

    def run(self) -> None:
        # update filepath, it will be in running jobs now
        if self._config.batch_size == 0:
            # on batch size zero we only call the callable with the filename
            self.call_callable_with_items(self._file_path)
            # save progress
            commit_load_package_state()
        else:
            current_index = self._destination_state.get(self._storage_id, 0)
            for batch in self.get_batches(current_index):
                self.call_callable_with_items(batch)
                current_index += len(batch)
                self._destination_state[self._storage_id] = current_index
                # save progress
                commit_load_package_state()

    def call_callable_with_items(self, items: TDataItems) -> None:
        if not items:
            return
        # call callable
        if self._callable_requires_job_client_args:
            self._callable(items, self._load_table, job_client=self._job_client)  # type: ignore
        else:
            self._callable(items, self._load_table)

    @abstractmethod
    def get_batches(self, start_index: int) -> Iterable[TDataItems]:
        pass


class DestinationParquetLoadJob(DestinationLoadJob):
    def get_batches(self, start_index: int) -> Iterable[TDataItems]:
        # stream items
        from dlt.common.libs.pyarrow import pyarrow

        # guard against changed batch size after restart of loadjob
        assert (
            start_index % self._config.batch_size
        ) == 0, "Batch size was changed during processing of one load package"

        # on record batches we cannot drop columns, we need to
        # select the ones we want to keep
        keep_columns = list(self._load_table["columns"].keys())
        start_batch = start_index / self._config.batch_size
        with pyarrow.parquet.ParquetFile(self._file_path) as reader:
            for record_batch in reader.iter_batches(
                batch_size=self._config.batch_size, columns=keep_columns
            ):
                if start_batch > 0:
                    start_batch -= 1
                    continue
                yield record_batch


class DestinationJsonlLoadJob(DestinationLoadJob):
    def get_batches(self, start_index: int) -> Iterable[TDataItems]:
        current_batch: TDataItems = []

        # stream items
        with FileStorage.open_zipsafe_ro(self._file_path) as f:
            for line in f:
                encoded_json = json.typed_loads(line)
                if isinstance(encoded_json, dict):
                    encoded_json = [encoded_json]

                for item in encoded_json:
                    # find correct start position
                    if start_index > 0:
                        start_index -= 1
                        continue
                    # skip internal columns
                    for column in self._skipped_columns:
                        item.pop(column, None)
                    current_batch.append(item)
                    if len(current_batch) == self._config.batch_size:
                        yield current_batch
                        current_batch = []
                yield current_batch



import dataclasses
from typing import Any, ClassVar, Dict, Final, List, Optional

from dlt.common.configuration import configspec
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.configuration.specs import AwsCredentials
from dlt.destinations.impl.athena.utils import is_s3_tables_catalog


DEFAULT_AWS_DATA_CATALOG = "awsdatacatalog"
DEFAULT_TABLE_LOCATION_LAYOUT = "{dataset_name}/{table_name}"


@configspec
class LakeformationConfig:
    enabled: Optional[bool] = None
    tags: Optional[Dict[str, str]] = None


@configspec
class AthenaClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="athena", init=False, repr=False, compare=False)  # type: ignore[misc]
    query_result_bucket: Optional[str] = None
    credentials: AwsCredentials = None
    athena_work_group: Optional[str] = None
    aws_data_catalog: str = DEFAULT_AWS_DATA_CATALOG
    staging_aws_data_catalog: Optional[str] = None
    connection_params: Optional[Dict[str, Any]] = None
    force_iceberg: Optional[bool] = None
    table_location_layout: Optional[str] = DEFAULT_TABLE_LOCATION_LAYOUT
    table_properties: Optional[Dict[str, str]] = None
    lakeformation_config: Optional[LakeformationConfig] = None
    info_tables_query_threshold: int = 90
    # athena slows down when this value is too high, see for context:
    # https://github.com/dlt-hub/dlt/issues/2529
    db_location: Optional[str] = None

    __config_gen_annotations__: ClassVar[List[str]] = [
        "query_result_bucket",
        "athena_work_group",
        "aws_data_catalog",
        "info_tables_query_threshold",
    ]

    def to_connector_params(self, use_catalog_name: bool = True) -> Dict[str, Any]:
        native_credentials = self.credentials.to_native_representation()
        return {
            "s3_staging_dir": self.query_result_bucket,
            "work_group": self.athena_work_group,
            "catalog_name": self.aws_data_catalog if use_catalog_name else None,
            **(self.connection_params or {}),
            **native_credentials,
        }

    def _is_s3_tables_catalog(self) -> bool:
        return is_s3_tables_catalog(self.aws_data_catalog)

    def __str__(self) -> str:
        """Return displayable destination location"""
        if self.staging_config:
            return f"{self.staging_config} on {self.aws_data_catalog}"
        else:
            return "[no staging set]"



from functools import partial
from typing import Any, List

import sqlglot.expressions as sge
from sqlglot.schema import Schema as SQLGlotSchema

from dlt.common.destination.capabilities import TCasefoldIdentifier
from dlt.common.libs.sqlglot import bind_query
from dlt.destinations.sql_client import SqlClientBase


def _normalize_query(
    qualified_query: sge.Query,
    sqlglot_schema: SQLGlotSchema,
    *,
    sql_client: SqlClientBase[Any],
    casefold_identifier: TCasefoldIdentifier,
) -> sge.Query:
    """Backward compatibility wrapper for bind_query.

    TODO: remove after next dlthub release
    """
    return bind_query(
        qualified_query,
        sqlglot_schema,
        expand_table_name=partial(
            sql_client.make_qualified_table_name_path, quote=False, casefold=False
        ),
        casefold_identifier=casefold_identifier,
    )


def build_row_counts_expr(
    table_name: str,
    quoted_identifiers: bool = True,
    dlt_load_id_col: str = None,
    load_id: str = None,
) -> sge.Select:
    """Builds a SQL expression to count rows in a table, optionally filtering by a load ID."""

    if bool(load_id) ^ bool(dlt_load_id_col):
        raise ValueError("Both `load_id` and `dlt_load_id_col` must be provided together.")

    table_expr = sge.Table(this=sge.to_identifier(table_name, quoted=quoted_identifiers))

    select_expr = sge.Select(
        expressions=[
            sge.Alias(
                this=sge.Literal.string(table_name),
                alias=sge.to_identifier("table_name"),
            ),
            sge.Alias(
                this=sge.Count(this=sge.Star()),
                alias=sge.to_identifier("row_count"),
            ),
        ]
    ).from_(table_expr)

    if load_id:
        where_condition = sge.EQ(
            this=sge.to_identifier(dlt_load_id_col, quoted=quoted_identifiers),
            expression=sge.Literal.string(load_id),
        )
        select_expr = select_expr.where(where_condition)

    return select_expr


def build_select_expr(
    table_name: str,
    selected_columns: List[str] = None,
    quoted_identifiers: bool = True,
) -> sge.Select:
    """Builds a SQL expression to select all or selected rows from a table."""

    if selected_columns is None:
        selected_columns = ["*"]

    table_expr = sge.Table(this=sge.to_identifier(table_name, quoted=quoted_identifiers))

    columns_expr = [
        (
            sge.Star()
            if col == "*"
            else sge.Column(this=sge.to_identifier(col, quoted=quoted_identifiers))
        )
        for col in selected_columns
    ]

    select_expr = sge.Select(expressions=columns_expr).from_(table_expr)

    return select_expr



"""SQL database source helpers"""

import warnings
from typing import (
    Callable,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Iterator,
    Type,
    Union,
)
import operator
from abc import ABC, abstractmethod

import dlt
from dlt.common import logger
from dlt.common.configuration.specs import (
    BaseConfiguration,
    ConnectionStringCredentials,
    configspec,
)
from dlt.common.exceptions import (
    DltException,
    MissingDependencyException,
    ValueErrorWithKnownValues,
)
from dlt.common.schema import TTableSchemaColumns
from dlt.common.schema.typing import TWriteDispositionDict
from dlt.common.schema.utils import merge_columns
from dlt.common.typing import TColumnNames, TDataItem, TSortOrder, add_value_to_literal
from dlt.common.jsonpath import extract_simple_field_name
from dlt.common.utils import is_typeerror_due_to_wrong_call

from dlt.extract import Incremental
from dlt.extract.items_transform import LimitItem

from .arrow_helpers import row_tuples_to_arrow
from .schema_types import (
    default_table_adapter,
    Table,
    SelectAny,
    ReflectionLevel,
    TTypeAdapter,
    table_to_resource_hints,
)

from dlt.common.libs.sql_alchemy import (
    Engine,
    CompileError,
    create_engine,
    MetaData,
    sa,
    TextClause,
    ORACLE_NUMBER,
    OracleDialect,
)

TableBackend = Literal["sqlalchemy", "pyarrow", "pandas", "connectorx"]
SelectClause = Union[SelectAny, TextClause]
TQueryAdapter = Union[
    Callable[[SelectAny, Table], SelectClause],
    Callable[[SelectAny, Table, Incremental[Any], Engine], SelectClause],
]
TTableAdapter = Callable[[Table], Optional[Union[SelectAny, Table]]]

TABLE_LOADER_REGISTRY: Dict[str, Type["BaseTableLoader"]] = {}
"""Maps backend name to the table loader class that handles it."""


def _maybe_fix_0000_timezone(df: Any) -> Any:
    """Optionally convert +00:00 timezone to UTC on Arrow tables."""
    try:
        from dlt.common.libs.pyarrow import set_plus0000_timezone_to_utc, pyarrow

        # TODO: skip when Arrow releases timezone fix
        if isinstance(df, pyarrow.Table):
            return set_plus0000_timezone_to_utc(df)
    except MissingDependencyException:
        pass
    return df


class BaseTableLoader(ABC):
    """Base class for SQL table loaders.

    Provides query building infrastructure including incremental loading support.
    Subclasses must implement ``load_rows`` to handle the actual data retrieval.
    For loaders that use a SQLAlchemy connection, subclass ``TableLoader`` instead
    to reuse the result-conversion helpers.

    Args:
        engine: SQLAlchemy ``Engine`` used for query compilation and (optionally)
            execution.
        backend: Requested output format (``"sqlalchemy"``, ``"pyarrow"``,
            ``"pandas"``, ``"connectorx"``).
        table: Reflected SQLAlchemy ``Table`` object.
        columns: dlt column schema hints for the table.
        chunk_size: Number of rows per batch.
        incremental: Optional incremental loading state.
        query_adapter_callback: Optional callback to modify the generated query.
        limit: Optional row/time limit from the resource.
    """

    def __init__(
        self,
        engine: Engine,
        backend: TableBackend,
        table: Table,
        columns: TTableSchemaColumns,
        chunk_size: int = 1000,
        incremental: Optional[Incremental[Any]] = None,
        query_adapter_callback: Optional[TQueryAdapter] = None,
        limit: Optional[LimitItem] = None,
    ) -> None:
        self.engine = engine
        self.backend = backend
        self.table = table
        self.columns = columns
        self.chunk_size = chunk_size
        self.query_adapter_callback = query_adapter_callback
        self.incremental = incremental
        self.limit = limit
        if incremental:
            column_name = extract_simple_field_name(incremental.cursor_path)

            if column_name is None:
                raise ValueError(
                    f"Cursor path `{incremental.cursor_path}` must be a simple column name (e.g."
                    " `created_at`)"
                )

            try:
                self.cursor_column = table.c[column_name]
            except KeyError as e:
                raise KeyError(
                    f"Cursor column `{incremental.cursor_path}` does not exist in table"
                    f" `{table.name}`"
                ) from e
            self.last_value = incremental.last_value
            self.end_value = incremental.end_value
            self.row_order: TSortOrder = self.incremental.row_order
            self.on_cursor_value_missing = self.incremental.on_cursor_value_missing
            self.range_start = self.incremental.range_start
            self.range_end = self.incremental.range_end
        else:
            self.cursor_column = None
            self.last_value = None
            self.end_value = None
            self.row_order = None
            self.on_cursor_value_missing = None
            self.range_start = None
            self.range_end = None

    def _make_query(self) -> SelectAny:
        table = self.table
        query = table.select()

        # use limit step to limit the query
        if self.limit:
            # limit works with chunks (by default)
            limit = self.limit.limit(self.chunk_size)
            if limit is not None:
                query = query.limit(limit)

        if not self.incremental:
            return query  # type: ignore[no-any-return]
        last_value_func = self.incremental.last_value_func

        # generate where
        if last_value_func is max:  # Query ordered and filtered according to last_value function
            filter_op = operator.ge if self.range_start == "closed" else operator.gt
            filter_op_end = operator.lt if self.range_end == "open" else operator.le
        elif last_value_func is min:
            filter_op = operator.le if self.range_start == "closed" else operator.lt
            filter_op_end = operator.gt if self.range_end == "open" else operator.ge
        else:  # Custom last_value, load everything and let incremental handle filtering
            return query  # type: ignore[no-any-return]

        where_clause = True
        if self.last_value is not None:
            where_clause = filter_op(self.cursor_column, self.last_value)
            if self.end_value is not None:
                where_clause = sa.and_(
                    where_clause, filter_op_end(self.cursor_column, self.end_value)
                )

            if self.on_cursor_value_missing == "include":
                where_clause = sa.or_(where_clause, self.cursor_column.is_(None))
        if self.on_cursor_value_missing == "exclude":
            where_clause = sa.and_(where_clause, self.cursor_column.isnot(None))

        if where_clause is not True:
            query = query.where(where_clause)

        # generate order by from declared row order
        order_by = None
        if (self.row_order == "asc" and last_value_func is max) or (
            self.row_order == "desc" and last_value_func is min
        ):
            order_by = self.cursor_column.asc()
        elif (self.row_order == "asc" and last_value_func is min) or (
            self.row_order == "desc" and last_value_func is max
        ):
            order_by = self.cursor_column.desc()
        if order_by is not None:
            query = query.order_by(order_by)

        return query  # type: ignore[no-any-return]

    def make_query(self) -> SelectClause:
        if self.query_adapter_callback:
            try:
                return self.query_adapter_callback(  # type: ignore[call-arg]
                    self._make_query(), self.table, self.incremental, self.engine
                )
            except TypeError as type_err:
                if not is_typeerror_due_to_wrong_call(type_err, self.query_adapter_callback):
                    raise
                return self.query_adapter_callback(  # type: ignore[call-arg]
                    self._make_query(), self.table
                )

        return self._make_query()

    def compile_query(self, query: SelectClause) -> str:
        """Compile a SQLAlchemy query into a SQL string with literal binds.

        Useful for backends that execute raw SQL strings (ConnectorX, ADBC, etc.).

        Raises:
            NotImplementedError: When the query cannot be compiled to a string.
        """
        try:
            return str(query.compile(self.engine, compile_kwargs={"literal_binds": True}))
        except CompileError as ex:
            raise NotImplementedError(
                f"Query for table `{self.table.name}` could not be compiled to string."
                f" If you are on SQLAlchemy 1.4.x, upgrade to 2.x: `{ex}`"
            ) from ex

    def get_connection_url(self) -> str:
        """Return a plain database connection URL derived from the engine.

        Strips the SQLAlchemy driver portion (e.g. ``+psycopg2``) so the URL
        is suitable for non-SQLAlchemy backends such as ConnectorX or ADBC.
        Override this method when a backend requires a different connection
        string format (e.g. MSSQL ODBC → go-mssqldb conversion).
        """
        return self.engine.url._replace(
            drivername=self.engine.url.get_backend_name()
        ).render_as_string(hide_password=False)

    @abstractmethod
    def load_rows(self, backend_kwargs: Dict[str, Any] = None) -> Iterator[TDataItem]:
        """Load rows from the table and yield them as data items.

        Args:
            backend_kwargs: Backend-specific keyword arguments passed through
                from the resource configuration.

        Yields:
            Data items in the format determined by the loader implementation
            (dicts, Arrow tables, DataFrames, etc.).
        """
        ...


class TableLoader(BaseTableLoader):
    """Default table loader using a SQLAlchemy connection.

    Supports ``"sqlalchemy"``, ``"pyarrow"`` and ``"pandas"`` backends.
    Override ``_load_rows`` to customise query execution (e.g. pagination)
    while reusing ``_convert_result`` for backend format conversion.
    """

    def load_rows(self, backend_kwargs: Dict[str, Any] = None) -> Iterator[TDataItem]:
        # make copy of kwargs
        backend_kwargs = dict(backend_kwargs or {})
        query = self.make_query()
        yield from self._load_rows(query, backend_kwargs)

    def _load_rows(
        self, query: SelectClause, backend_kwargs: Dict[str, Any]
    ) -> Iterator[TDataItem]:
        """Execute *query* over a SQLAlchemy connection and yield converted results.

        Override this method to change the execution strategy (e.g. to add
        pagination or custom retry logic).  Use ``_convert_result`` to convert
        each ``CursorResult`` into the appropriate backend format.
        """
        with self.engine.connect() as conn:
            result = conn.execution_options(yield_per=self.chunk_size).execute(query)
            try:
                yield from self._convert_result(result, backend_kwargs)
            finally:
                result.close()

    def _convert_result(self, result: Any, backend_kwargs: Dict[str, Any]) -> Iterator[TDataItem]:
        """Convert a ``CursorResult`` into data items for the configured backend.

        Partitions *result* into chunks of ``chunk_size`` and converts each
        partition according to ``self.backend``:

        * ``"sqlalchemy"`` -- list of dicts
        * ``"pandas"`` -- pandas ``DataFrame``
        * ``"pyarrow"`` -- Arrow table
        """
        # NOTE: cursor returns not normalized column names! may be quite useful
        # in case of Oracle dialect that normalizes columns
        columns = list(result.keys())
        for partition in result.partitions(size=self.chunk_size):
            if self.backend == "sqlalchemy":
                yield [dict(row._mapping) for row in partition]
            elif self.backend == "pandas":
                from dlt.common.libs.pandas_sql import _wrap_result

                df = _wrap_result(
                    partition,
                    columns,
                    **{"dtype_backend": "pyarrow", **backend_kwargs},
                )
                yield df
            elif self.backend == "pyarrow":
                # add columns present in cursor but missing from schema hints
                for col_name in columns:
                    if col_name not in self.columns:
                        self.columns[col_name] = {"name": col_name}
                # build arrow table using cursor column order for correct
                # positional mapping (self.columns may be in schema order
                # after merge_columns reordering)
                cursor_columns = {c: self.columns[c] for c in columns}
                table = row_tuples_to_arrow(
                    partition,
                    columns=cursor_columns,
                    tz=backend_kwargs.get("tz", "UTC"),
                )
                yield table


class ConnectorXTableLoader(BaseTableLoader):
    """Table loader using ConnectorX for data retrieval. Yields Arrow tables."""

    def load_rows(self, backend_kwargs: Dict[str, Any] = None) -> Iterator[TDataItem]:
        # make copy of kwargs
        backend_kwargs = dict(backend_kwargs or {})
        query = self.make_query()
        yield from self._load_rows_connectorx(query, backend_kwargs)

    def _load_rows_connectorx(
        self, query: SelectClause, backend_kwargs: Dict[str, Any]
    ) -> Iterator[TDataItem]:
        try:
            import connectorx as cx
        except ImportError:
            raise MissingDependencyException("Connector X table backend", ["connectorx"])

        import pyarrow as pa
        from dlt.common.libs.pyarrow import cast_date64_columns_to_timestamp

        # default settings
        backend_kwargs = {
            "protocol": "binary",
            **backend_kwargs,
        }

        is_streaming = False
        if "return_type" in backend_kwargs:
            if backend_kwargs["return_type"] == "arrow_stream":
                is_streaming = True
                backend_kwargs["batch_size"] = backend_kwargs.get("batch_size", self.chunk_size)
        else:
            backend_kwargs["return_type"] = "arrow"

        conn = backend_kwargs.pop("conn", self.get_connection_url())
        query_str = self.compile_query(query)
        logger.info(f"Executing query on ConnectorX: {query_str}")

        if is_streaming:
            record_reader = cx.read_sql(conn, query_str, **backend_kwargs)
            for record_batch in record_reader:
                table = pa.Table.from_batches((record_batch,), schema=record_batch.schema)
                yield cast_date64_columns_to_timestamp(_maybe_fix_0000_timezone(table))
        else:
            df = cx.read_sql(conn, query_str, **backend_kwargs)
            if len(df) > self.chunk_size:
                logger.info(
                    f"The size of the dataset being loaded is more than {self.chunk_size},"
                    " consider using streaming mode (see"
                    " https://dlthub.com/docs/dlt-ecosystem/verified-sources/sql_database"
                    "/configuration#connectorx)"
                )
            yield _maybe_fix_0000_timezone(df)


# populate built-in backends
TABLE_LOADER_REGISTRY.update(
    {
        "sqlalchemy": TableLoader,
        "pyarrow": TableLoader,
        "pandas": TableLoader,
        "connectorx": ConnectorXTableLoader,
    }
)


def register_table_loader_backend(
    backend_name: str,
    loader_class: Type[BaseTableLoader],
) -> None:
    """Register a custom table loader backend.

    After registration the *backend_name* can be used as the ``backend``
    argument of ``sql_table`` / ``sql_database``.

    Args:
        backend_name: Name for the backend (used in ``backend`` parameter).
        loader_class: A subclass of ``BaseTableLoader`` that handles data loading.
    """
    if not issubclass(loader_class, BaseTableLoader):
        raise ValueError(
            f"Invalid table loader: `{loader_class.__name__}`. "
            "Must be a subclass of `BaseTableLoader`."
        )
    TABLE_LOADER_REGISTRY[backend_name] = loader_class
    add_value_to_literal(TableBackend, backend_name)


def get_table_loader_class(backend_name: str) -> Type[BaseTableLoader]:
    """Look up the table loader class for the given backend name."""
    try:
        return TABLE_LOADER_REGISTRY[backend_name]
    except KeyError:
        raise ValueErrorWithKnownValues("backend", backend_name, list(TABLE_LOADER_REGISTRY.keys()))


def table_rows(
    engine: Engine,
    table: Union[Table, str],
    metadata: MetaData,
    chunk_size: int,
    backend: TableBackend,
    incremental: Optional[Incremental[Any]],
    table_adapter_callback: TTableAdapter,
    reflection_level: ReflectionLevel,
    backend_kwargs: Dict[str, Any],
    type_adapter_callback: Optional[TTypeAdapter],
    included_columns: Optional[List[str]],
    excluded_columns: Optional[List[str]],
    query_adapter_callback: Optional[TQueryAdapter],
    resolve_foreign_keys: bool,
    table_loader_class: Optional[Type[BaseTableLoader]] = None,
) -> Iterator[TDataItem]:
    resource = None
    limit = None
    try:
        resource = dlt.current.resource()
        limit = resource.limit
        resource_columns_hints_require_data = callable(resource.columns)
        if resource_columns_hints_require_data and backend == "pyarrow":
            table_name = table.name if isinstance(table, Table) else table
            logger.warning(
                f"Dynamic column hints for '{table_name}' cannot be applied with pyarrow "
                "backend. Use static hints (dict/list) to override reflected types."
            )
    except DltException:
        # in old versions of dlt, resource is not available, so we need to reflect the table again
        pass
    if isinstance(table, str):  # Reflection is deferred
        table = Table(
            table,
            metadata,
            autoload_with=engine,
            extend_existing=True,
            resolve_fks=resolve_foreign_keys,
        )
        table = _execute_table_adapter(
            table, table_adapter_callback, included_columns, excluded_columns
        )
        hints = table_to_resource_hints(
            table,
            reflection_level,
            type_adapter_callback,
            backend == "sqlalchemy",  # skip nested types
            resolve_foreign_keys=resolve_foreign_keys,
        )

        # set the primary_key in the incremental
        # TODO: check for primary key in resource._hints
        if incremental and incremental.primary_key is None:
            primary_key = hints["primary_key"]
            if primary_key is not None:
                incremental.primary_key = primary_key

        # Merge resource hints with reflection hints before yielding
        if resource and hints.get("columns") and not callable(resource.columns):
            hints["columns"] = merge_columns(hints["columns"], resource.columns)

        # yield empty record to set hints and create schema
        # Note: Empty list [] will be written as typed-jsonl (object format), but actual
        # data rows will be written in their native format (e.g., parquet for arrow backend).
        yield dlt.mark.with_hints(
            [],
            dlt.mark.make_hints(
                **hints,
            ),
        )
        # Set columns_hints for TableLoader
        columns_hints = hints["columns"]
    else:
        # table was already reflected -> try to use resource hints
        if not resource or callable(resource.columns):
            hints = table_to_resource_hints(
                table,
                reflection_level,
                type_adapter_callback,
                backend == "sqlalchemy",  # skip nested types
                resolve_foreign_keys=resolve_foreign_keys,
            )
            columns_hints = hints["columns"]

        else:
            # take column hints from resource (which includes user applied hints)
            # Handle callable columns hint (can't resolve without data item)
            columns_hints = resource.columns

    # warn about and remove columns_hints entries that don't match any
    # reflected source column.
    source_columns = {c.name for c in table.columns}
    unmatched = [c for c in columns_hints if c not in source_columns]
    if unmatched:
        logger.warning(
            "Resource columns %s for table '%s' do not match any reflected"
            " source columns and will be ignored. Note: source column names"
            " are not normalized at this stage. Use the original database"
            " column names in apply_hints.",
            unmatched,
            table.name,
        )
        columns_hints = {c: columns_hints[c] for c in columns_hints if c not in unmatched}

    # determine loader class from registry
    if table_loader_class is None:
        table_loader_class = get_table_loader_class(backend)

    loader = table_loader_class(
        engine,
        backend,
        table,
        columns_hints,
        incremental=incremental,
        chunk_size=chunk_size,
        limit=limit,
        query_adapter_callback=query_adapter_callback,
    )
    try:
        yield from loader.load_rows(backend_kwargs)
    finally:
        # dispose the engine if created for this particular table
        # NOTE: database wide engines are not disposed, not externally provided
        if getattr(engine, "may_dispose_after_use", False):
            engine.dispose()


def engine_from_credentials(
    credentials: Union[ConnectionStringCredentials, Engine, str],
    may_dispose_after_use: bool = False,
    **backend_kwargs: Any,
) -> Engine:
    if isinstance(credentials, Engine):
        return credentials
    if isinstance(credentials, ConnectionStringCredentials):
        credentials = credentials.to_native_representation()
    engine = create_engine(credentials, **backend_kwargs)
    setattr(engine, "may_dispose_after_use", may_dispose_after_use)  # noqa
    return engine  # type: ignore[no-any-return]


def unwrap_json_connector_x(field: str) -> TDataItem:
    """Creates a transform function to be added with `add_map` that will unwrap JSON columns
    ingested via connectorx. Such columns are additionally quoted and translate SQL NULL to json "null"
    """
    import pyarrow.compute as pc
    import pyarrow as pa

    def _unwrap(table: TDataItem) -> TDataItem:
        col_index = table.column_names.index(field)
        # remove quotes
        column = table[field]  # pc.replace_substring_regex(table[field], '"(.*)"', "\\1")
        # convert json null to null
        column = pc.replace_with_mask(
            column,
            pc.equal(column, "null").combine_chunks(),
            pa.scalar(None, pa.string()),
        )
        return table.set_column(col_index, table.schema.field(col_index), column)

    return _unwrap


def remove_nullability_adapter(table: Table) -> Table:
    """A table adapter that removes nullability from columns."""
    for col in table.columns:
        # subqueries may not have nullable attr
        if hasattr(col, "nullable"):
            col.nullable = None
    return table


def _execute_table_adapter(
    table: Table,
    adapter: Optional[TTableAdapter],
    included_columns: Optional[List[str]],
    excluded_columns: Optional[List[str]],
) -> Table:
    """Executes default table adapter on `table` and then `adapter` if defined"""
    default_table_adapter(table, included_columns, excluded_columns)
    if adapter:
        # backward compat: old adapters do not return a value
        maybe_query = adapter(table)
        if maybe_query is not None:
            # here we ignore that returned table may be a Select (subquery)
            # otherwise typing gets really complicated
            # TODO: maybe type that later
            table = maybe_query  # type: ignore[assignment]

    return table


def _detect_precision_hints_deprecated(value: Optional[bool]) -> None:
    if value is None:
        return

    msg = (
        "`detect_precision_hints` argument is deprecated and will be removed in a future release. "
    )
    if value:
        msg += "Use `reflection_level='full_with_precision'` which has the same effect instead."

    warnings.warn(
        msg,
        DeprecationWarning,
    )


@configspec
class SqlTableResourceConfiguration(BaseConfiguration):
    credentials: Union[ConnectionStringCredentials, Engine, str] = None
    table: str = None
    schema: Optional[str] = None
    incremental: Optional[Incremental] = None  # type: ignore[type-arg]
    chunk_size: int = 50000
    backend: TableBackend = "sqlalchemy"
    detect_precision_hints: Optional[bool] = None
    defer_table_reflect: Optional[bool] = False
    reflection_level: Optional[ReflectionLevel] = "full"
    included_columns: Optional[List[str]] = None
    excluded_columns: Optional[List[str]] = None
    write_disposition: Optional[TWriteDispositionDict] = None
    primary_key: Optional[TColumnNames] = None
    merge_key: Optional[TColumnNames] = None
    engine_kwargs: Optional[Dict[str, Any]] = None


def _oracle_column_reflect_listener(
    inspector: Any, table: Any, column_info: Dict[str, Any]
) -> None:
    """
    SQLAlchemy event listener for `column_reflect` that enforces Oracle NUMBER type
    to translate to python decimal.Decimal.

    Oracle NUMBER may express floating- or fixed-point numbers, but floats
    don't conform to IEEE754 standard, so we're always using "decimal" type
    to preserve values as accurately as possible. SQLAlchemy2 uses different
    logic for determining asdecimal, we're overriding it here.
    """
    column_type = column_info.get("type")
    if isinstance(column_type, ORACLE_NUMBER):
        column_info["type"] = ORACLE_NUMBER(
            precision=column_type.precision,
            scale=column_type.scale,
            asdecimal=True,
        )


def default_engine_adapter_callback(engine: Engine, metadata: MetaData) -> None:
    """Applies default engine adaptations for known dialects.

    For Oracle dialect, registers an event listener on the provided MetaData that forces
    NUMBER columns to be reflected as Python Decimal to preserve numeric precision.

    Args:
        engine: The SQLAlchemy engine to check dialect for.
        metadata: The MetaData instance to register the listener on.
    """
    if isinstance(engine.dialect, OracleDialect):
        sa.event.listen(metadata, "column_reflect", _oracle_column_reflect_listener)



import inspect
from typing import Any, Type, Optional, Callable, Union, overload
from typing_extensions import Concatenate
from dlt.common.destination.reference import DestinationReference, AnyDestination, Destination
from dlt.common.reflection.spec import get_spec_name_from_f
from dlt.common.typing import AnyFun

from functools import wraps

from dlt.common import logger
from dlt.common.destination import TLoaderFileFormat
from dlt.common.typing import TDataItems
from dlt.common.schema import TTableSchema
from dlt.common.destination.capabilities import TLoaderParallelismStrategy

from dlt.common.utils import get_callable_name, get_full_callable_name, is_inner_callable
from dlt.destinations.impl.destination.factory import destination as _destination
from dlt.destinations.impl.destination.configuration import (
    TDestinationCallableParams,
    CustomDestinationClientConfiguration,
)


@overload
def destination(
    func: None = ...,
    /,
    *,
    loader_file_format: TLoaderFileFormat = None,
    batch_size: int = 10,
    name: str = None,
    naming_convention: str = "direct",
    skip_dlt_columns_and_tables: bool = True,
    max_table_nesting: int = 0,
    spec: Type[CustomDestinationClientConfiguration] = None,
    max_parallel_load_jobs: Optional[int] = None,
    loader_parallelism_strategy: Optional[TLoaderParallelismStrategy] = None,
) -> Callable[
    [Callable[Concatenate[Union[TDataItems, str], TTableSchema, TDestinationCallableParams], Any]],
    Callable[TDestinationCallableParams, _destination],
]:
    """A decorator that transforms a function into a custom destination with configuration parameters.

    This overload handles parameterized decorator usage where configuration is provided
    in the decorator call, and the function to be decorated is applied later.

    #### Example Usage:

    >>> @dlt.destination(batch_size=100, loader_file_format="parquet")
    >>> def my_destination(items, table, api_url: str = dlt.config.value, api_secret = dlt.secrets.value):
    >>>     print(table["name"])
    >>>     print(items)
    >>>
    >>> p = dlt.pipeline("chess_pipeline", destination=my_destination)

    Args:
        func (None): Must be None for this overload. The actual function will be provided
            when the decorator is applied.

        loader_file_format (TLoaderFileFormat, optional): Defines the format in which files are stored
            in the load package before being sent to the destination function. Defaults to None.

        batch_size (int, optional): Defines how many items per function call are batched together
            and sent as an array. If set to 0, instead of passing actual data items, you will receive
            one call per load job with the path of the file as the "items" argument. Defaults to 10.

        name (str, optional): Defines the name of the destination. If not provided, defaults to
            the name of the decorated function.

        naming_convention (str, optional): Controls how table and column names are normalized.
            The default value, "direct", keeps all names unchanged.

        skip_dlt_columns_and_tables (bool, optional): Defines whether internal dlt tables and columns
            are included in the custom destination function. Defaults to True.

        max_table_nesting (int, optional): Defines how deep the normalizer will go to flatten nested
            fields in your data to create subtables. This overrides any source settings and defaults
            to 0, meaning no nested tables are created.

        spec (Type[CustomDestinationClientConfiguration], optional): Defines a configuration spec used
            to inject arguments into the decorated function. Arguments not included in the spec will
            not be injected. Defaults to None.

        max_parallel_load_jobs (Optional[int], optional): Defines the maximum number of load jobs
            that can run concurrently during the load process. Defaults to None.

        loader_parallelism_strategy (Optional[TLoaderParallelismStrategy], optional): Determines the
            load job parallelism strategy. Can be "sequential" (equivalent to max_parallel_load_jobs=1),
            "table-sequential" (one load job per table at a time), or "parallel". Defaults to None.

    Returns:
        Callable[[Callable[Concatenate[Union[TDataItems, str], TTableSchema, TDestinationCallableParams], Any]], Callable[TDestinationCallableParams, _destination]]:
            A decorator function that accepts the destination function and returns a callable that can
            be used to create a custom dlt destination instance.
    """
    ...


@overload
def destination(
    func: Callable[
        Concatenate[Union[TDataItems, str], TTableSchema, TDestinationCallableParams], Any
    ],
    /,
    *,
    loader_file_format: TLoaderFileFormat = None,
    batch_size: int = 10,
    name: str = None,
    naming_convention: str = "direct",
    skip_dlt_columns_and_tables: bool = True,
    max_table_nesting: int = 0,
    spec: Type[CustomDestinationClientConfiguration] = None,
    max_parallel_load_jobs: Optional[int] = None,
    loader_parallelism_strategy: Optional[TLoaderParallelismStrategy] = None,
) -> Callable[TDestinationCallableParams, _destination]:
    """Creates a destination factory from a function by directly passing it as an argument.

    This overload handles direct function calls where the destination function is passed
    as the first argument, rather than using decorator syntax.

    #### Example Usage:

    >>> def my_destination(items, table, api_url: str = dlt.config.value, api_secret = dlt.secrets.value):
    >>>     print(table["name"])
    >>>     print(items)
    >>>
    >>> dest = dlt.destination(my_destination, batch_size=100, loader_file_format="parquet")
    >>> p = dlt.pipeline("chess_pipeline", destination=dest)

    Args:
        func (Callable[Concatenate[Union[TDataItems, str], TTableSchema, TDestinationCallableParams], Any]):
            A callable that takes two positional arguments "items" and "table",
            followed by any number of keyword arguments with default values. This function
            will process the incoming data and does not need to return anything. The keyword
            arguments can represent configuration or secret values.

        loader_file_format (TLoaderFileFormat): Defines the format in which files are stored in the load package
            before being sent to the destination function.

        batch_size (int): Defines how many items per function call are batched together and sent as an array.
            If set to 0, instead of passing actual data items, you will receive one call per load job
            with the path of the file as the "items" argument. You can then open and process that file as needed.

        name (str): Defines the name of the destination created by the destination decorator.
            Defaults to the name of the function.

        naming_convention (str): Controls how table and column names are normalized.
            The default value, "direct", keeps all names unchanged.

        skip_dlt_columns_and_tables (bool): Defines whether internal dlt tables and columns
            are included in the custom destination function. Defaults to True.

        max_table_nesting (int): Defines how deep the normalizer will go to flatten nested fields
            in your data to create subtables. This overrides any source settings and defaults to 0,
            meaning no nested tables are created.

        spec (Type[CustomDestinationClientConfiguration]): Defines a configuration spec used
            to inject arguments into the decorated function. Arguments not included in the spec will not be injected.

        max_parallel_load_jobs (Optional[int]): Defines the maximum number of load jobs
            that can run concurrently during the load process.

        loader_parallelism_strategy (Optional[TLoaderParallelismStrategy]): Determines the load job parallelism strategy.
            Can be "sequential" (equivalent to max_parallel_load_jobs=1), "table-sequential" (one load job per table at a time),
            or "parallel".

    Returns:
        Callable[TDestinationCallableParams, _destination]: A callable that can be used to create a custom dlt destination instance.
    """
    ...


@overload
def destination(
    destination_name: str,
    /,
    *,
    destination_type: Optional[str] = None,
    credentials: Optional[Any] = None,
    **kwargs: Any,
) -> AnyDestination:
    """Instantiates a destination from the provided destination name and type by retrieving the corresponding
    destination factory and initializing it with the given credentials and keyword arguments.

    Args:
        destination_name (str): The name of the destination instance to initialize.

        destination_type (Optional[str]): The type of the destination to instantiate.

        credentials (Optional[Any]): Credentials used to connect to the destination.
            May be an instance of a credential class supported by the respective destination.

        **kwargs (Any): Additional keyword arguments passed to the destination factory.

    Returns:
        AnyDestination: An initialized destination.
    """
    ...


def destination(
    func_or_name: Union[Optional[AnyFun], str] = None,
    /,
    *,
    loader_file_format: TLoaderFileFormat = None,
    batch_size: int = 10,
    name: str = None,
    naming_convention: str = "direct",
    skip_dlt_columns_and_tables: bool = True,
    max_table_nesting: int = 0,
    spec: Type[CustomDestinationClientConfiguration] = None,
    max_parallel_load_jobs: Optional[int] = None,
    loader_parallelism_strategy: Optional[TLoaderParallelismStrategy] = None,
    **kwargs: Any,
) -> Any:
    """When used as a decorator, transforms a function that takes two positional arguments, "items" and "table",
    along with any number of keyword arguments with default values, into a callable that will create a custom destination.
    The function itself does not return anything, the keyword arguments can represent configuration or secret values.

    When used as a function with the first argument being a string (the destination name),
    instantiates a destination from the provided destination name and type by retrieving the corresponding
    destination factory and initializing it with the given credentials and keyword arguments.
    """

    def decorator(
        destination_callable: Callable[
            Concatenate[Union[TDataItems, str], TTableSchema, TDestinationCallableParams], Any
        ]
    ) -> Callable[TDestinationCallableParams, _destination]:
        # resolve destination name
        destination_name = name or get_callable_name(destination_callable)

        # synthesize new Destination factory
        class _ConcreteDestinationBase(_destination):
            def __init__(self, **kwargs: Any):
                default_args = dict(
                    spec=spec,
                    destination_callable=destination_callable,
                    loader_file_format=loader_file_format,
                    batch_size=batch_size,
                    destination_name=destination_name,
                    naming_convention=naming_convention,
                    skip_dlt_columns_and_tables=skip_dlt_columns_and_tables,
                    max_table_nesting=max_table_nesting,
                    max_parallel_load_jobs=max_parallel_load_jobs,
                    loader_parallelism_strategy=loader_parallelism_strategy,
                )
                super().__init__(**{**default_args, **kwargs})

        cls_name = get_spec_name_from_f(destination_callable, kind="Destination")
        module = inspect.getmodule(destination_callable)
        # synthesize type
        D: Type[_destination] = type(
            cls_name,
            (_ConcreteDestinationBase,),
            {"__module__": module.__name__, "__orig_base__": _destination},
        )
        # add to the module
        setattr(module, cls_name, D)
        # register only standalone destinations, no inner
        if not is_inner_callable(destination_callable):
            DestinationReference.register(D, get_full_callable_name(destination_callable))

        @wraps(destination_callable)
        def wrapper(
            *args: TDestinationCallableParams.args, **kwargs: TDestinationCallableParams.kwargs
        ) -> _destination:
            if args:
                logger.warning(
                    "Ignoring positional arguments for destination callable %s",
                    destination_callable,
                )
            return D(**kwargs)  # type: ignore[arg-type]

        setattr(wrapper, "_factory", D)  # noqa
        return wrapper

    if func_or_name is None:
        # we're called with parens.
        return decorator
    elif not isinstance(func_or_name, str):
        # we're called as @dlt.destination without parens.
        return decorator(func_or_name)

    # Factory mode: create destination instance from given string
    destination_type = kwargs.pop("destination_type", None)
    if destination_type:
        destination = Destination.from_reference(
            ref=destination_type,
            destination_name=func_or_name,
            **kwargs,
        )
    elif kwargs.get("destination_callable"):
        destination = Destination.from_reference(
            ref="destination",
            destination_name=func_or_name,
            **kwargs,
        )
    else:
        destination = Destination.from_reference(
            ref=func_or_name,
            **kwargs,
        )

    return destination



from types import TracebackType
from typing import Optional, Type, Iterable, cast, List

from dlt.destinations.job_impl import FinalizedLoadJob
from dlt.common.destination.client import LoadJob, PreparedTableSchema, JobClientBase
from dlt.common.typing import AnyFun
from dlt.common.storages.load_package import destination_state
from dlt.common.storages.load_storage import ParsedLoadJobFileName
from dlt.common.configuration import create_resolved_partial

from dlt.common.schema import Schema, TSchemaTables
from dlt.common.schema.utils import is_dlt_table_or_column
from dlt.common.destination import DestinationCapabilitiesContext

from dlt.destinations.impl.destination.configuration import CustomDestinationClientConfiguration
from dlt.destinations.job_impl import (
    DestinationJsonlLoadJob,
    DestinationParquetLoadJob,
)


class DestinationClient(JobClientBase):
    """Sink Client"""

    def __init__(
        self,
        schema: Schema,
        config: CustomDestinationClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        config.ensure_callable()
        super().__init__(schema, config, capabilities)
        self.config: CustomDestinationClientConfiguration = config
        # create pre-resolved callable to avoid multiple config resolutions during execution of the jobs
        self.destination_callable = create_resolved_partial(
            cast(AnyFun, self.config.destination_callable), self.config
        )

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        pass

    def is_storage_initialized(self) -> bool:
        return True

    def drop_storage(self) -> None:
        pass

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        return super().update_stored_schema(only_tables, expected_update)

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        # skip internal tables and remove columns from schema if so configured
        if self.config.skip_dlt_columns_and_tables:
            if is_dlt_table_or_column(table["name"], self.schema._dlt_tables_prefix):
                return FinalizedLoadJob(file_path)

        skipped_columns: List[str] = []
        if self.config.skip_dlt_columns_and_tables:
            for column in list(self.schema.get_table(table["name"])["columns"].keys()):
                if is_dlt_table_or_column(column, self.schema._dlt_tables_prefix):
                    skipped_columns.append(column)

        # save our state in destination name scope
        load_state = destination_state()

        parsed_file = ParsedLoadJobFileName.parse(file_path)
        if parsed_file.file_format == "parquet":
            return DestinationParquetLoadJob(
                file_path,
                self.config,
                load_state,
                self.destination_callable,
                skipped_columns,
            )
        if parsed_file.file_format in ["jsonl", "typed-jsonl"]:
            return DestinationJsonlLoadJob(
                file_path,
                self.config,
                load_state,
                self.destination_callable,
                skipped_columns,
            )
        return None

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        table = super().prepare_load_table(table_name)
        if self.config.skip_dlt_columns_and_tables:
            for column in list(table["columns"].keys()):
                if is_dlt_table_or_column(column, self.schema._dlt_tables_prefix):
                    table["columns"].pop(column)
        return table

    def complete_load(self, load_id: str) -> None: ...

    def __enter__(self) -> "DestinationClient":
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        pass



"""Configuration for Fabric Warehouse destination - extends Synapse configuration with COPY INTO support"""

from typing import Optional, Final, ClassVar, Dict, Any, List
from dlt.common.configuration import configspec
from dlt.common.configuration.specs import AzureServicePrincipalCredentials
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.exceptions import MissingDependencyException
from dlt import version

_AZURE_STORAGE_EXTRA = f"{version.DLT_PKG_NAME}[az]"


@configspec(init=False)
class FabricCredentials(AzureServicePrincipalCredentials):
    """Credentials for Microsoft Fabric Warehouse with Service Principal authentication.

    Fabric Warehouse requires Azure AD Service Principal authentication.
    Inherits from AzureServicePrincipalCredentials for Service Principal fields and
    automatic fallback to DefaultAzureCredential.
    """

    drivername: str = "mssql+pyodbc"
    """SQLAlchemy driver name for SQL Server/Fabric."""

    host: str = None
    """Fabric Warehouse host (e.g., abc12345-6789-def0-1234-56789abcdef0.datawarehouse.fabric.microsoft.com)"""

    port: int = 1433
    """Database port (default: 1433)"""

    database: str = None
    """Fabric Warehouse database name"""

    connect_timeout: int = 15
    """Connection timeout in seconds (default: 15)"""

    # Override to make optional - not needed for Fabric Warehouse credentials (only for staging)
    azure_storage_account_name: Optional[str] = None
    """Not used for Fabric Warehouse credentials (only staging credentials need this)"""

    def on_partial(self) -> None:
        """Enable fallback to DefaultAzureCredential if explicit credentials not provided."""
        try:
            from azure.identity import DefaultAzureCredential
        except ModuleNotFoundError:
            raise MissingDependencyException(self.__class__.__name__, [_AZURE_STORAGE_EXTRA])

        # If no explicit Service Principal credentials, use default credentials
        if not self.azure_client_id or not self.azure_client_secret or not self.azure_tenant_id:
            self._set_default_credentials(DefaultAzureCredential())
            # Resolve if we have warehouse connection details (not storage account name)
            if self.host and self.database:
                self.resolve()

    def get_odbc_dsn_dict(self) -> Dict[str, Any]:
        """Build ODBC DSN dictionary with Fabric-specific settings."""
        params = {
            "DRIVER": "{ODBC Driver 18 for SQL Server}",
            "SERVER": f"{self.host},{self.port}",
            "DATABASE": self.database,
            "AUTHENTICATION": "ActiveDirectoryServicePrincipal",
            "LongAsMax": "yes",  # Required for UTF-8 collation support
            "Encrypt": "yes",
            "TrustServerCertificate": "no",
        }

        # Add Service Principal credentials if provided
        if self.azure_client_id and self.azure_tenant_id and self.azure_client_secret:
            params["UID"] = f"{self.azure_client_id}@{self.azure_tenant_id}"
            params["PWD"] = str(self.azure_client_secret)

        return params

    def to_odbc_dsn(self) -> str:
        """Build ODBC connection string for pyodbc."""
        params = self.get_odbc_dsn_dict()
        return ";".join(f"{k}={v}" for k, v in params.items())

    def to_native_credentials(self) -> Optional[Any]:
        """Return credentials in a format suitable for the native driver/library."""
        return self.get_odbc_dsn_dict()


@configspec
class FabricClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    """Configuration for Fabric Warehouse destination with staging and collation support.

    Uses FabricCredentials for Service Principal authentication.
    Supports OneLake/Lakehouse or Azure Blob Storage staging with COPY INTO for efficient data loading.

    Example usage with OneLake/Lakehouse staging (recommended):
        fabric(
            credentials={
                "host": "abc12345-6789-def0-1234-56789abcdef0.datawarehouse.fabric.microsoft.com",
                "database": "mydb",
                "tenant_id": "your-tenant-id",
                "client_id": "your-client-id",
                "client_secret": "your-client-secret",
            },
            staging=filesystem(
                # IMPORTANT: Must use workspace GUID and lakehouse GUID (not names)
                # Format: abfss://<workspace_guid>@onelake.dfs.fabric.microsoft.com/<lakehouse_guid>/Files
                bucket_url="abfss://12345678-1234-1234-1234-123456789012@onelake.dfs.fabric.microsoft.com/87654321-4321-4321-4321-210987654321/Files",
                # IMPORTANT: Must specify Service Principal credentials (same as warehouse)
                credentials={
                    "azure_storage_account_name": "onelake",
                    "azure_account_host": "onelake.blob.fabric.microsoft.com",
                    "azure_tenant_id": "your-tenant-id",
                    "azure_client_id": "your-client-id",
                    "azure_client_secret": "your-client-secret",
                },
            ),
            collation="Latin1_General_100_BIN2_UTF8",
        )

    Note: The bucket_url must use GUIDs for both workspace and lakehouse, not their display names.
    You can find these GUIDs in the Fabric portal workspace/lakehouse URLs.

    Example usage with Azure Blob Storage staging:
        fabric(
            credentials={
                "host": "abc12345-6789-def0-1234-56789abcdef0.datawarehouse.fabric.microsoft.com",
                "database": "mydb",
                "tenant_id": "your-tenant-id",
                "client_id": "your-client-id",
                "client_secret": "your-client-secret",
            },
            staging=filesystem(
                bucket_url="az://your-container",
                credentials={
                    "azure_storage_account_name": "your-account-name",
                    "azure_storage_account_key": "your-account-key",
                },
            ),
            collation="Latin1_General_100_BIN2_UTF8",
        )
    """

    destination_type: Final[str] = "fabric"  # type: ignore[misc]
    credentials: Optional[FabricCredentials] = None

    collation: Optional[str] = "Latin1_General_100_BIN2_UTF8"
    """Database collation to use for text columns.

    Note: Fabric Warehouse does not support table indexing. Storage is automatically managed by the system.
    """

    # Set to False by default because PRIMARY KEY and UNIQUE constraints
    # are NOT ENFORCED in Fabric and can lead to inaccurate results
    create_indexes: bool = False
    """Whether `primary_key` and `unique` column hints are applied."""

    has_case_sensitive_identifiers: bool = True
    """Whether identifiers (table/column names) are case-sensitive. Depends on database collation."""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "default_table_index_type",
        "create_indexes",
    ]

    """Database collation for varchar columns. Fabric supports:
    - Latin1_General_100_BIN2_UTF8 (default, case-sensitive)
    - Latin1_General_100_CI_AS_KS_WS_SC_UTF8 (case-insensitive)

    Both have UTF-8 encoding. LongAsMax=yes is automatically configured.
    """


__all__ = ["FabricCredentials", "FabricClientConfiguration"]



"""Module with mark functions that make data to be specially processed"""
from typing import Any, Dict, List, Union

from dlt.extract import (
    make_hints,
    make_nested_hints,
    with_file_import,
    with_hints,
    with_table_name,
    materialize_schema_item as materialize_table_schema,
)


def with_json_flatten(
    columns: Dict[str, Union[bool, List[str]]],
    keep_original: bool = False,
) -> Dict[str, Dict[str, Any]]:
    """
    Helper function to create column hints for JSON expansion.
    
    Args:
        columns: Dict mapping column names to flatten specification
            - True: Flatten entire JSON object
            - List[str]: Flatten only specified paths (e.g., ["user.name", "user.email"])
        keep_original: Whether to keep original JSON string column alongside flattened columns
    
    Returns:
        Column hints dict compatible with apply_hints() or make_hints(columns=...)
    
    Example:
        @dlt.resource(
            apply_hints=dlt.mark.with_json_flatten(
                {"metadata": True, "data": ["user.name", "user.email"]},
                keep_original=True
            )
        )
        def my_resource():
            yield {"metadata": '{"name": "John"}'}
    """
    hints = {}
    for col, spec in columns.items():
        hints[col] = {
            "x-json-flatten": spec,
            "x-json-keep-original": keep_original,
        }
    return hints


__all__ = [
    "with_table_name",
    "with_hints",
    "with_file_import",
    "make_hints",
    "make_nested_hints",
    "materialize_table_schema",
    "with_json_flatten",
]



import marimo

__generated_with = "0.19.2"
app = marimo.App(width="medium")

with app.setup:
    from typing import Any, cast
    from itertools import chain

    import marimo as mo

    import dlt
    from dlt.common.utils import without_none

    pipeline_name = None


@app.cell
def _():
    mo.stop(pipeline_name is None)
    pipeline = dlt.attach(pipeline_name)
    return (pipeline,)


@app.function
def pipeline_details(pipeline: dlt.Pipeline) -> list[dict]:
    """
    Get the details of a pipeline.
    """
    try:
        credentials = str(pipeline.dataset().destination_client.config.credentials)
    except Exception:
        credentials = "Could not resolve credentials"

    details_dict = {
        "Pipeline name": pipeline.pipeline_name,
        "Destination": (
            pipeline.destination.destination_description
            if pipeline.destination
            else "No destination set"
        ),
        "Credentials": credentials,
        "Dataset name": pipeline.dataset_name,
        "Schema name": (
            pipeline.default_schema_name
            if pipeline.default_schema_name
            else "No default schema set"
        ),
        "Pipeline working dir": pipeline.working_dir,
    }

    return [{"Key": k, "Value": v} for k, v in details_dict.items()]


@app.cell
def _():
    def _align_dict_keys(items: list[dict]) -> list[dict]:
        """
        Makes sure all dicts have the same keys, sets "-" as default. Makes for nicer rendering in marimo table
        """
        items = cast(list[dict], [without_none(i) for i in items])
        all_keys = set(chain.from_iterable(i.keys() for i in items))
        for i in items:
            i.update({key: "-" for key in all_keys if key not in i})
        return items

    def style_cell(row_id: str, name: str, __: Any) -> dict[str, str]:
        """
        Style a cell in a table.

        Args:
            row_id (str): The id of the row.
            name (str): The name of the column.
            __ (Any): The value of the cell.

        Returns:
            Dict[str, str]: The css style of the cell.
        """
        style = {}  # "background-color": "white" if (int(row_id) % 2 == 0) else "#f4f4f9"}
        if name.lower() == "name":
            style["font-weight"] = "bold"
        return style

    @mo.cache
    def create_table_list(
        pipeline: dlt.Pipeline,
        show_internals: bool = False,
        show_child_tables: bool = True,
    ) -> list[dict[str, str]]:
        """Create a list of tables for the pipeline.

        Args:
            pipeline_name (str): The name of the pipeline to create the table list for.
        """
        FILL_VALUE = "-"

        # get tables and filter as needed
        tables = list(pipeline.default_schema.tables.values())
        if not show_child_tables:
            tables = [t for t in tables if t.get("parent") is None]

        table_list = []
        keys = ("name", "parent", "resource", "write_disposition", "description")
        for table in tables:
            info = {key.capitalize().replace("_", " "): table.get(key, FILL_VALUE) for key in keys}
            table_list.append(info)

        table_list.sort(key=lambda x: str(x["Name"]))
        if not show_internals:
            table_list = [t for t in table_list if not str(t["Name"]).lower().startswith("_dlt")]

        return _align_dict_keys(table_list)

    return create_table_list, style_cell


@app.cell
def _(dlt_schema_table_list, pipeline):
    schema_dict = pipeline.default_schema.to_dict().copy()
    schema_dict["tables"] = {
        col_name: col
        for col_name, col in schema_dict["tables"].items()
        if col_name in [table["Name"].lower() for table in dlt_schema_table_list.value]
    }

    mo.mermaid(dlt.Schema.from_dict(schema_dict).to_mermaid())
    return


@app.cell
def _(create_table_list, pipeline, style_cell):
    table_list = create_table_list(
        pipeline,
        show_internals=True,
        show_child_tables=True,
    )

    dlt_schema_table_list = mo.ui.table(
        table_list,  # type: ignore[arg-type]
        style_cell=style_cell,
        initial_selection=list(range(0, len(table_list))),
    )
    dlt_schema_table_list
    return (dlt_schema_table_list,)


@app.cell
def _(pipeline):
    raw_schema = mo.lazy(
        mo.json(pipeline.default_schema.to_dict(), label=pipeline.default_schema_name)
    )

    raw_schema_yaml = mo.lazy(
        mo.ui.code_editor(
            pipeline.default_schema.to_pretty_yaml(),
            language="yaml",
        )
    )

    raw_schema_json = mo.lazy(
        mo.ui.code_editor(
            pipeline.default_schema.to_pretty_json(),
            language="json",
        )
    )

    raw_schema_dbml = mo.lazy(
        mo.ui.code_editor(
            value=pipeline.default_schema.to_dbml(),
        )
    )

    raw_schema_dot = mo.lazy(
        mo.ui.code_editor(
            value=pipeline.default_schema.to_dot(),
        )
    )

    raw_schema_format_tabs = mo.ui.tabs(
        {
            "dict": raw_schema,
            "JSON": raw_schema_json,
            "YAML": raw_schema_yaml,
            "DBML": raw_schema_dbml,
            "DOT": raw_schema_dot,
        },
    )
    raw_schema_format_tabs
    return


if __name__ == "__main__":
    app.run()



"""A module that imports all pluggy plugins implementations to work with pyproject entrypoint"""
from typing import Any, Dict, Optional

from dlt.common.configuration import plugins as _plugins
from dlt.common.configuration.specs.pluggable_run_context import RunContextBase
from dlt.common.runtime.run_context import RunContext

# import workspace plugins before core plugin
from dlt._workspace._plugins import *  # noqa
from dlt._workspace.cli._plugins import *  # noqa


# make sure OSS context is tried last
@_plugins.hookimpl(specname="plug_run_context", trylast=True)
def plug_run_context_impl(
    run_dir: Optional[str], runtime_kwargs: Optional[Dict[str, Any]]
) -> Optional[RunContextBase]:
    return RunContext(run_dir)
